

from lantz import ArgumentParserSC


def main(args=None):
    """Run simulators.
    """
    parser = ArgumentParserSC('demo', CHOICES, description='Run Lantz QtDemo.')
    parser.dispatch(args)


def testpanel_demo(args=None):
    """Run simulators.
    """

    from lantz.qt.utils.qt import QtGui
    """
    QtGui.QMessageBox.information(None,
                                  'Lantz Test Panel Demo'
                                  'Please run the following command in the console\n'
                                  'to start the simulator:\n\n'
                                  '    lantz sims fungen tcp')
    """
    print('Please make sure that the simulator is running.\n'
          'You can start it by running the following command in another terminal:\n\n'
          'lantz sims fungen tcp')

    from lantz.drivers.examples import LantzSignalGenerator

    from lantz.qt import start_test_app, wrap_driver_cls

    QLantzSignalGenerator = wrap_driver_cls(LantzSignalGenerator)
    with QLantzSignalGenerator('TCPIP::localhost::5678::SOCKET') as inst:
        start_test_app(inst)


def uifile_demo(args=None):
    print('Please make sure that the simulator is running.\n'
          'You can start it by running the following command in another terminal:\n\n'
          'lantz sims fungen tcp')

    import os

    UIFILE = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'assets', 'fungen.ui')

    from lantz.drivers.examples import LantzSignalGenerator

    from lantz.qt import start_gui, wrap_driver_cls

    QLantzSignalGenerator = wrap_driver_cls(LantzSignalGenerator)
    with QLantzSignalGenerator('TCPIP::localhost::5678::SOCKET') as inst:
        start_gui(UIFILE, inst)


def featscan_demo(args=None):

    print('Please make sure that the simulator is running.\n'
          'You can start it by running the following command in another terminal:\n\n'
          'lantz sims fungen tcp')

    # We import a helper function to start the app
    from lantz.qt import start_gui_app, wrap_driver_cls

    # The block consists of two parts the backend and the frontend
    from lantz.qt.blocks import FeatScan, FeatScanUi

    # An this you know already
    from lantz.drivers.examples import LantzSignalGenerator

    QLantzSignalGenerator = wrap_driver_cls(LantzSignalGenerator)

    with QLantzSignalGenerator('TCPIP::localhost::5678::SOCKET') as fungen:

        # Here we instantiate the backend setting 'frequency' as the Feat to scan
        # and specifying in which instrument
        app = FeatScan('frequency', instrument=fungen)

        # Now we use the helper to start the app.
        # It takes a Backend instance and a FrontEnd class
        start_gui_app(app, FeatScanUi)


def flock_demo(args=None):

    # TODO: Make this an actual flock

    print('Please make sure that the simulator is running.\n'
          'You can start it by running the following command in another terminal:\n\n'
          'lantz sims fungen tcp')

    # We import a helper function to start the app
    from lantz.qt import start_gui_app, wrap_driver_cls

    # The block consists of two parts the backend and the frontend
    from lantz.qt.blocks import FeatScan, FeatScanUi

    # An this you know already
    from lantz.drivers.examples import LantzSignalGenerator

    QLantzSignalGenerator = wrap_driver_cls(LantzSignalGenerator)

    with QLantzSignalGenerator('TCPIP::localhost::5678::SOCKET') as fungen:

        # Here we instantiate the backend setting 'frequency' as the Feat to scan
        # and specifying in which instrument
        app = FeatScan('frequency', instrument=fungen)

        # Now we use the helper to start the app.
        # It takes a Backend instance and a FrontEnd class
        start_gui_app(app, FeatScanUi)



CHOICES = {'testpanel': testpanel_demo,
           'uifile': uifile_demo,
           'featscan': featscan_demo}

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
    lantz.qt
    ~~~~~~~~

    Implements UI functionality for lantz using Qt.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from .app import start_test_app, start_gui, start_gui_app, Backend, Frontend, InstrumentSlot
from .objwrapper import wrap_driver_cls
from .utils.qt import QtCore, QtGui, SuperQObject, MetaQObject



# -*- coding: utf-8 -*-
"""
    lantz.ui.featscan
    ~~~~~~~~~~~~~~~~~

    A Feat Scan frontend and Backend. Builds upon Scan.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from ..utils.qt import QtCore
from ..widgets import WidgetMixin
from ..app import start_gui_app, InstrumentSlot

from . import Scan, ScanUi


class FeatScan(Scan):
    """A backend to scan a feat for a given instrument."""

    #: Signal emitted before starting a new iteration
    #: Parameters: loop counter, step value, overrun
    iteration = QtCore.Signal(int, int, bool)

    #: Signal emitted when the loop finished.
    #: The parameter is used to inform if the loop was canceled.
    loop_done = QtCore.Signal(bool)

    instrument = InstrumentSlot

    #: Name of the scanned feat
    #: :type: str

    def __init__(self, feat_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.feat_name = feat_name

    def _pre_body(self, counter, new_value, overrun):
        setattr(self.instrument, self.feat_name, new_value)

    @property
    def feat_units(self):
        """Units of the scanned feat."""
        target = self.instrument
        feat_name = self.feat_name
        feat = target.feats[feat_name]
        return str(feat.units)


class FeatScanUi(ScanUi):
    """A Frontend displaying scan parameters with appropriate units."""

    def connect_backend(self):
        target = self.backend.instrument
        feat_name = self.backend.feat_name

        feat = target.feats[feat_name]

        def _pimp(widget):
            WidgetMixin.wrap(widget)
            widget.bind_feat(feat)

        for name in 'start stop step_size'.split():
            _pimp(getattr(self.widget, name))

        if feat.limits and len(feat.limits) == 3:
            self.widget.step_size.setMinimum(feat.limits[2])
        else:
            self.widget.step_size.setMinimum(0)
        super().connect_backend()


if __name__ == '__main__':
    from lantz.drivers.examples import LantzSignalGenerator

    with LantzSignalGenerator('TCPIP::localhost::5678::SOCKET') as inst:
        app = FeatScan('frequency', fungen=inst)
        start_gui_app(app, FeatScanUi)



# -*- coding: utf-8 -*-
"""
    lantz.widgets.dialog_units
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Dialog to allow the user to select a new, compatible unit.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from lantz.core import Q_
from ..utils.qt import QtCore, QtGui


class UnitInputDialog(QtGui.QDialog):
    """Dialog to select new units. Checks compatibility while typing
    and does not allow to continue if incompatible.

    Parameters
    ----------
    units : Quantity
        Quantity with source units.
    parent : PyQt Widget
        parent widget.


    Example
    -------
    >>> new_units = UnitInputDialog.get_units('ms')

    """

    def __init__(self, units, parent=None):
        super().__init__(parent)
        self.setupUi(parent)
        self.units = units
        self.source_units.setText(str(units.units))

    def setupUi(self, parent):
        self.resize(275, 172)
        self.setWindowTitle('Convert units')
        self.layout = QtGui.QVBoxLayout(parent)
        self.layout.setSizeConstraint(QtGui.QLayout.SetFixedSize)
        align = (QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing | QtCore.Qt.AlignVCenter)

        self.layout1 = QtGui.QHBoxLayout()
        self.label1 = QtGui.QLabel()
        self.label1.setMinimumSize(QtCore.QSize(100, 0))
        self.label1.setText('Convert from:')
        self.label1.setAlignment(align)

        self.layout1.addWidget(self.label1)
        self.source_units = QtGui.QLineEdit()
        self.source_units.setReadOnly(True)
        self.layout1.addWidget(self.source_units)

        self.layout.addLayout(self.layout1)

        self.layout2 = QtGui.QHBoxLayout()
        self.label2 = QtGui.QLabel()
        self.label2.setMinimumSize(QtCore.QSize(100, 0))
        self.label2.setText('to:')
        self.label2.setAlignment(align)
        self.layout2.addWidget(self.label2)

        self.destination_units = QtGui.QLineEdit()
        self.layout2.addWidget(self.destination_units)

        self.layout.addLayout(self.layout2)

        self.message = QtGui.QLabel()
        self.message.setText('')
        self.message.setAlignment(QtCore.Qt.AlignCenter)

        self.layout.addWidget(self.message)

        self.buttonBox = QtGui.QDialogButtonBox()
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Ok)
        self.layout.addWidget(self.buttonBox)
        self.buttonBox.setEnabled(False)

        self.buttonBox.accepted.connect(self.accept)
        self.destination_units.textChanged.connect(self.check)

        self.setLayout(self.layout)
        self.destination_units.setFocus()

    def check(self):
        units = self.destination_units.text().strip()
        if not units:
            return
        try:
            new_units = Q_(1, units)
            factor = self.units.to(new_units).magnitude
        except LookupError or SyntaxError:
            self.message.setText('Cannot parse units')
            self.buttonBox.setEnabled(False)
        except ValueError:
            self.message.setText('Incompatible units')
            self.buttonBox.setEnabled(False)
        except AttributeError:
            self.message.setText('Unknown units')
            self.buttonBox.setEnabled(False)
        else:
            self.message.setText('factor {:f}'.format(factor))
            self.buttonBox.setEnabled(True)

    @staticmethod
    def get_units(units):
        """Creates and display a UnitInputDialog and return new units.

        Parameters
        ----------
        units : str
            current units.

        Returns
        -------
        str or None
            output compatible units.
            Returns None if cancelled.

        """

        if isinstance(units, str):
            units = Q_(1, units)

        dialog = UnitInputDialog(Q_(1, units.units))
        if dialog.exec_():
            return dialog.destination_units.text()
        return None


# -*- coding: utf-8 -*-
"""
    lantz.utils
    ~~~~~~~~~~~

    A package with utility modules.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import os

# Indicates if the documentation is being built in the current process.
# :type: bool
LANTZ_BUILDING_DOCS = os.environ.get('LANTZ_BUILDING_DOCS', 'False') == 'True'


# -*- coding: utf-8 -*-
"""
    lantz.widgets.numeric
    ~~~~~~~~~~~~~~~~~~~~~

    Numeric widgets.
    - QDoubleSpinBox
    - CheckBox
    - QLineEdit
    - QSlider
    - QDial
    - QProgressBar
    - QScrollBar
    - QLCDNumber

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz.core import Q_
from lantz.core.helpers import MISSING, UNSET

from ..utils.qt import QtGui
from .dialog_units import UnitInputDialog
from .common import WidgetMixin, register_wrapper


@register_wrapper
class MagnitudeMixin(WidgetMixin):

    _WRAPPED = (QtGui.QDoubleSpinBox, QtGui.QSpinBox)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if self._units and event.text() == 'u':
            self.change_units(request_new_units(self.value()))

    def bind_feat(self, feat):
        super().bind_feat(feat)

        #: self._units are the current units displayed by the widget.
        #: Respects units declared in the suffix

        if feat.units:
            suf = (self.suffix() if hasattr(self, 'suffix') else feat.units) or feat.units
            self._units = Q_(1, suf)
            self.change_units(self._units)
        else:
            self._units = None
            if feat.limits:
                self.change_limits(None)

    def change_units(self, new_units):
        """Update displayed suffix and stored units.
        """
        if new_units is None:
            return
        try:
            rescaled = self.value().to(new_units)
        except ValueError:
            # incompatible units
            return None
        else:
            if hasattr(self, 'setSuffix'):
               self.setSuffix(' ' + str(new_units.units))

            self.change_limits(new_units)
            self._units = new_units

            self.setValue(rescaled)

    def change_limits(self, new_units):
        """Change the limits (range) of the control taking the original values
        from the feat and scaling them to the new_units.
        """
        if not hasattr(self, 'setRange'):
            return

        rng = self._feat.limits

        if not rng:
            return

        if new_units:
            conv = lambda ndx: Q_(rng[ndx], self._feat.units).to(new_units).magnitude
        else:
            conv = lambda ndx: rng[ndx]

        if len(rng) == 1:
            self.setRange(0, conv(0))
        else:
            self.setRange(conv(0), conv(1))
            if len(rng) == 3:
                self.setSingleStep(conv(2))

    def value(self):
        """Get widget value and scale by units."""
        if self._units:
            return Q_(super().value(), self._units)
        return super().value()

    def setValue(self, value):
        """Set widget value scaled by units.
        """
        if value is MISSING or value is UNSET:
            font = QtGui.QFont()
            font.setItalic(True)
            self.setFont(font)
        elif isinstance(value, Q_):
            super().setValue(value.to(self._units).magnitude)
        else:
            super().setValue(value)


@register_wrapper
class SliderMixin(MagnitudeMixin):

    _WRAPPED = (QtGui.QSlider, QtGui.QDial, QtGui.QProgressBar, QtGui.QScrollBar)

    def setReadOnly(self, value):
        super().setEnabled(not value)


@register_wrapper
class LCDNumberMixin(MagnitudeMixin):

    _WRAPPED = (QtGui.QLCDNumber, )

    @classmethod
    def _wrap(cls, widget):
        super()._wrap(widget)
        #TODO: Create a real valueChanged Signal.
        widget.valueChanged = widget.overflow

    def setReadOnly(self, value):
        super().setEnabled(not value)

    def setValue(self, value):
        if value is MISSING or value is UNSET:
            font = QtGui.QFont()
            font.setItalic(True)
            self.setFont(font)
            return
        elif isinstance(value, Q_):
            super().display(value.to(self._units).magnitude)
        else:
            super().display(value)

    def value(self):
        return super().value()


def request_new_units(current_units):
    """Ask for new units using a dialog box and return them.

    Parameters
    ----------
    current_units : Quantity
        current units or magnitude.

    Returns
    -------
    None or Quantity
    """
    new_units = UnitInputDialog.get_units(current_units)
    if new_units is None:
        return None

    try:
        return Q_(1, new_units)
    except LookupError:
        # cannot parse units
        return None




# -*- coding: utf-8 -*-
"""
    lantz.ui.layouts
    ~~~~~~~~~~~~~~~~

    Frontends to automatically locate widgets.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from ..utils.qt import QtGui
from ..app import Frontend


class _CommonUi:
    """The Loop frontend provides a GUI for the Rich Backend"""

    def _add(self, layout, parts):
        """Add widgets in parts to layout.

        Parameters
        ----------
        layout :
            
        parts :
            

        Returns
        -------

        """
        for part_name in parts:
            if part_name is True:
                layout.addStretch()
                continue

            part = getattr(self, part_name)

            if isinstance(part, Frontend):
                layout.addWidget(part)

            elif isinstance(part, tuple):
                # A tuple found in parts is considered nesting
                if isinstance(layout, self._inner):
                    sublayout = self._outer()
                elif isinstance(layout, self._outer):
                    sublayout = self._inner()
                else:
                    raise ValueError('Unknown parent layout %s' % layout)

                self._add(sublayout, part)

                layout.setLayout(sublayout)

            else:
                raise ValueError('Only Frontend or tuple are valid values '
                                 'valid for parts not %s (%s)' % (part, type(part)))

        return layout


class _PanelsUi(_CommonUi, Frontend):

    gui = 'placeholder.ui'

    auto_connect = False

    #: Tuple with the columns
    #: Each element can be:
    #:   - Frontend class: will be connected to the default backend.
    #:   - Front2Back(Frontend class, backend name): will be connect to a specific backend.
    #:   - tuple: will be iterated to obtain the rows.
    parts = ()

    _inner, _outer = None, None

    def setupUi(self):
        super().setupUi()
        layout = self._outer()
        self._add(layout, self.parts)
        self.widget.placeholder.setLayout(layout)


class VerticalUi(_PanelsUi):
    """Uses a vertical box layout to locate widgets."""

    _inner, _outer = QtGui.QHBoxLayout, QtGui.QVBoxLayout


class HorizonalUi(_PanelsUi):
    """Uses a horizontal box layout to locate widgets."""

    _inner, _outer = QtGui.QVBoxLayout, QtGui.QHBoxLayout


from ..app import Frontend


class ToolbarLeftRightUi(_CommonUi, Frontend):
    """The Loop frontend provides a GUI for the Rich Backend"""

    gui = 'left_main_center.ui'

    auto_connect = False

    #: Tuple with the columns
    #: Each element can be:
    #:   - Frontend class: will be connected to the default backend.
    #:   - Front2Back(Frontend class, backend name): will be connect to a specific backend.
    #:   - tuple: will be iterated to obtain the rows.
    toolbar = ()
    left = ()
    center = ()

    def setupUi(self):
        super().setupUi()

        self.widget.topbar.setLayout(self._add(QtGui.QHBoxLayout(), self.toolbar))
        self.widget.leftbar.setLayout(self._add(QtGui.QVBoxLayout(), self.left))
        self.setCentralWidget(self.center)




# -*- coding: utf-8 -*-
"""
    lantz.qt.chart
    ~~~~~~~~~~~~~~

    A chart frontend.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz.core import Q_

# Import Qt modules from lantz (pyside and pyqt compatible)
from ..utils.qt import QtGui

# These classes simplify application development
from ..app import Frontend


class ChartUi(Frontend):
    """A frontend with a x,y plot (powered by pyqtgraph)"""

    # a declarative way to indicate the user interface file to use.
    # The file must be located next to the python file where this class
    # is defined.
    gui = 'placeholder.ui'

    # connect widgets to instruments using connect_setup automatically.
    auto_connect = True

    _axis = {'x': 'bottom',
             'y': 'left'}

    def __init__(self, xlabel='', xunits='', ylabel='', yunits='', *args, **kwargs):
        self._labels = {'x': xlabel, 'y': ylabel}
        self._units = {'x': xunits, 'y': yunits}

        self._x = []
        self._y = []

        super().__init__(*args, **kwargs)

    @property
    def xlabel(self):
        """x-axis label."""
        return self._labels['x']

    @xlabel.setter
    def xlabel(self, value):
        self._labels['x'] = value
        self._relabel('x')

    @property
    def ylabel(self):
        """y-axis label."""
        return self._labels['y']

    @ylabel.setter
    def ylabel(self, value):
        self._labels['y'] = value
        self._relabel('y')

    @property
    def xunits(self):
        """x-axis units as a string."""
        return self._units['x']

    @xunits.setter
    def xunits(self, value):
        self._units['x'] = value
        self._relabel('x')

    @property
    def yunits(self):
        """y-axis units as a string."""
        return self._units['y']

    @yunits.setter
    def yunits(self, value):
        self._units['y'] = value
        self._relabel('y')

    def _relabel(self, axis):
        """Builds the actual label using the label and units for a given axis.
        Also builds a quantity to be used to normalized the data.

        Parameters
        ----------
        axis :
            x' or 'y'

        Returns
        -------

        """
        label = self._labels[axis]
        units = self._units[axis]
        if label and units:
            label = '%s [%s]' % (label, units)
        elif units:
            label = '[%s]' % units

        self.pw.setLabel(self._axis[axis], label)

        if units:
            setattr(self, '_q' + axis, Q_(1, units))

    def setupUi(self):
        import pyqtgraph as pg

        pg.setConfigOptions(antialias=True)
        # This method is called after gui has been loaded (referenced in self.widget)
        # to customize gui building. In this case, we are adding a plot widget.
        self.pw = pg.PlotWidget()

        self.curve = self.pw.plot(pen='y')
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.pw)
        self.widget.placeholder.setLayout(layout)

    def plot(self, x, y):
        """Add a pair of points to the plot.
        """
        x = x.to(self._qx).m
        y = y.to(self._qy).m
        self._x.append(x)
        self._y.append(y)
        self.curve.setData(self._x, self._y)

    def clear(self, *args):
        """Clear the plot.
        """
        self._x.clear()
        self._y.clear()



# -*- coding: utf-8 -*-
"""
    lantz.utils.qt
    ~~~~~~~~~~~~~~

    A Qt API selector that can be used to switch between PyQt and PySide.

    This uses the ETS 4.0 selection pattern of:
    PySide first, PyQt with API v2. second.

    Do not use this if you need PyQt with the old QString/QVariant API.

    Copied with modifications from the IPython Project.
    http://ipython.scipy.org/

    :copyright: IPython
    :license: BSD, see the IPython Project for more details.
"""

from .qt_loaders import (load_qt, QT_API_PYQT5, QT_API_PYSIDE2, QT_API_MOCK)
from .uic import build_loadUi

from ..config import QT_API

if QT_API:

    VALID = [QT_API_PYSIDE2, QT_API_PYQT5, QT_API_MOCK]

    if QT_API not in VALID:
        raise RuntimeError("Invalid Qt API %r, valid values are: %r or empty for system default." %
                           (QT_API, VALID))

    api_opts = [QT_API]
else:
    api_opts = [QT_API_PYQT5, ]

QtCore, QtGui, QtSvg, QT_API = load_qt(api_opts)

QtGui.loadUi, QtGui.loadUiType = build_loadUi(QT_API)

QtGui.QToolTip.setFont(QtGui.QFont('SansSerif', 10))


def superQ(QClass):
    """ Permits the use of super() in class hierarchies that contain Qt classes.
    Unlike QObject, SuperQObject does not accept a QObject parent. If it did,
    super could not be emulated properly (all other classes in the heierarchy
    would have to accept the parent argument--they don't, of course, because
    they don't inherit QObject.)
    This class is primarily useful for attaching signals to existing non-Qt
    classes. See QtKernelManagerMixin for an example.
    """
    class SuperQClass(QClass):

        def __new__(cls, *args, **kw):
            # We initialize QClass as early as possible. Without this, Qt complains
            # if SuperQClass is not the first class in the super class list.
            inst = QClass.__new__(cls)
            QClass.__init__(inst)
            return inst

        def __init__(self, *args, **kw):
            # Emulate super by calling the next method in the MRO, if there is one.
            mro = self.__class__.mro()
            for qt_class in QClass.mro():
                mro.remove(qt_class)
            next_index = mro.index(SuperQClass) + 1
            if next_index < len(mro):
                mro[next_index].__init__(self, *args, **kw)

    return SuperQClass


SuperQObject = superQ(QtCore.QObject)
MetaQObject = type(QtCore.QObject)



# -*- coding: utf-8 -*-
"""
    lantz.widgets.common
    ~~~~~~~~~~~~~~~~~~~~

    Common methods and classes to make PyQt widgets compatible with Lantz.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz.core.helpers import UNSET, MISSING

from lantz.core.feat import FeatProxy, DictFeatProxy, Feat, DictFeat
from ..log import LOGGER
from ..utils import LANTZ_BUILDING_DOCS
from ..utils.qt import QtGui


def register_wrapper(cls):
    """Register a class as lantz wrapper for QWidget subclasses.
    
    The class must contain a field (_WRAPPERS) with a tuple of the
    QWidget subclasses that it wraps.

    Parameters
    ----------

    Returns
    -------

    """
    for wrapped in cls._WRAPPED:
        if wrapped in cls._WRAPPERS:
            LOGGER.warn('{} is already registered to {}.'.format(wrapped, cls._WRAPPERS[wrapped]))

        if LANTZ_BUILDING_DOCS:
            cls._WRAPPERS[wrapped] = type(wrapped.__name__ + 'Wrapped',
                                          (cls, ), {'_IS_LANTZ_WRAPPER': True})
        else:
            cls._WRAPPERS[wrapped] = type(wrapped.__name__ + 'Wrapped',
                                          (cls, wrapped), {'_IS_LANTZ_WRAPPER': True})

    return cls


@register_wrapper
class WidgetMixin(object):
    """Mixin class to provide extra functionality to QWidget derived controls.
    
    Derived class must override _WRAPPED to indicate with which classes it
    can be mixed.
    
    To wrap an existing widget object use::
    
    
    If you want lantz to provide an appropriate wrapped widget for a given feat::
    
    
    In any case, after wrapping a widget you need to bind it to a feat::
    
    
    Finally, you need to

    Parameters
    ----------

    Returns
    -------

    >>> widget = QComboBox()
    >>> WidgetMixin.wrap(widget)
    
    >>> widget = WidgetMixin.from_feat(feat)
    
    >>> feat = driver.feats[feat_name]
    >>> widget.bind_feat(feat)
    
    >>> widget.lantz_target = driver
    """

    _WRAPPED = (QtGui.QWidget, )

    #: Dictionary linking Widget types with the function to patch them
    _WRAPPERS = {}

    def keyPressEvent(self, event):
        """When 'u' is pressed, request new units.
        When 'r' is pressed, get new value from the driver.
        """
        super().keyPressEvent(event)

        if event.text() == 'r':
            # This should also trigger a widget update if necessary.
            self.value_from_feat()

    def value(self):
        """Get widget value."""
        return super().value()

    def setValue(self, value):
        """Set widget value.
        """
        if value is MISSING:
            return
        super().setValue(value)

    def setReadOnly(self, value):
        """Set read only
        """
        super().setReadOnly(value)

    def value_from_feat(self):
        """Update the widget value with the current Feat value of the driver."""
        if self._feat is None or self._lantz_target is None:
            return

        if isinstance(self._feat, DictFeat):
            return getattr(self._lantz_target, self.feat.name)[self._feat_key]
        else:
            return getattr(self._lantz_target, self.feat.name)

    def value_to_feat(self):
        """Update the Feat value of the driver with the widget value."""
        if self._feat is None or self._lantz_target is None:
            return

        if isinstance(self._feat, DictFeat):
            getattr(self._lantz_target, self.feat.name)[self._feat_key] = self.value()
        else:
            setattr(self._lantz_target, self.feat.name, self.value())

    @property
    def readable(self):
        """If the Feat associated with the widget can be read (get)."""
        if self._feat is None:
            return False
        return self._feat.fget not in (None, MISSING)

    @property
    def writable(self):
        """If the Feat associated with the widget can be written (set)."""
        if self._feat is None:
            return False
        return self._feat.fset is not None

    def on_widget_value_changed(self, value, old_value=UNSET, key=MISSING):
        """When the widget is changed by the user, update the driver with
        the new value.
        """
        if key is not MISSING and key != self._feat_key:
            return
        if self._update_on_change:
            self.value_to_feat()

    def on_feat_value_changed(self, value, old_value=UNSET, key=MISSING):
        """When the driver value is changed, update the widget if necessary.
        """
        if key is not MISSING and key != self._feat_key:
            return
        if self.value() != value:
            self.setValue(value)

    @property
    def feat_key(self):
        """Key associated with the DictFeat."""
        return self._feat_key

    @feat_key.setter
    def feat_key(self, value):
        if self._lantz_target:
            getattr(self._lantz_target, self._feat.name + '_changed').disconnect(self.on_feat_value_changed)
        self._feat_key = value
        if self._lantz_target:
            getattr(self._lantz_target, self._feat.name + '_changed').connect(self.on_feat_value_changed)
        self.value_from_feat()

    @property
    def lantz_target(self):
        """Driver connected to the widget."""
        return self._lantz_target

    @lantz_target.setter
    def lantz_target(self, target):
        if self._lantz_target:
            self._feat_signal.disconnect(self.on_feat_value_changed)
            self.valueChanged.disconnect()

        if target:
            self._lantz_target = target

            self._feat_signal = getattr(self._lantz_target, self._feat.name + '_changed')
            self._feat_signal.connect(self.on_feat_value_changed)

            self.value_from_feat()
            self.valueChanged.connect(self.on_widget_value_changed)

    def bind_feat(self, feat):
        if isinstance(feat, (FeatProxy, DictFeatProxy)):
            self._feat = feat.proxied

        elif isinstance(feat, (Feat, DictFeat)):
            self._feat = feat

        else:
            raise ValueError('feat argument must be an instance of Feat, DictFeat, FeatProxy or DictFeatProxy, not %s' % feat.__class__)

        if isinstance(self._feat, DictFeat):
            keys = feat.keys
        else:
            keys = None

        if keys:
            self._feat_key = keys[0]
        else:
            self._feat_key = MISSING

        self.setReadOnly(not self.writable)

    @property
    def feat(self):
        return self._feat

    @classmethod
    def _wrap(cls, widget):
        ChildrenWidgets.patch(widget)
        widget._lantz_target = None
        widget._feat = None
        widget._update_on_change = True

    @classmethod
    def wrap(cls, widget):
        if hasattr(widget, '_lantz_wrapped'):
            return

        if getattr(widget, '_IS_LANTZ_WRAPPER', False):
            widget._wrap(widget)
        else:
            wrapper_class = cls._WRAPPERS.get(type(widget), cls)
            wrapper_class._wrap(widget)
            widget.__class__ = wrapper_class

        widget._lantz_wrapped = True

    @classmethod
    def from_feat(cls, feat, parent=None):
        """Return a widget appropriate to represent a lantz feature.

        Parameters
        ----------
        feat :
            a lantz feature proxy, the result of inst.feats[feat_name].
        parent :
            parent widget. (Default value = None)
        """

        _get = cls._WRAPPERS.get

        if feat.values:
            if isinstance(feat.values, dict):
                tmp = set(feat.values.keys())
            else:
                tmp = set(feat.values)

            if tmp == {True, False}:
                widget = _get(QtGui.QCheckBox)
            else:
                widget = _get(QtGui.QComboBox)
        elif not feat.units is None or feat.limits:
            widget = _get(QtGui.QDoubleSpinBox)
        else:
            widget= _get(QtGui.QLineEdit)

        widget = widget(parent)
        cls.wrap(widget)

        return widget


class ChildrenWidgets(object):
    """Convenience class to iterate children.

    Parameters
    ----------
    parent :
        parent widget.

    """

    def __init__(self, parent):
        self.parent = parent

    def __getattr__(self, item):
        return self.parent.findChild((QtGui.QWidget, ), item)

    def __iter__(self):
        pending = [self.parent, ]
        qualname = {self.parent: self.parent.objectName()}
        while pending:
            obj = pending.pop()
            for child in obj.children():
                if not isinstance(child, QtGui.QWidget):
                    continue
                qualname[child] = qualname[obj] + '.' + child.objectName()
                pending.append(child)
                yield child.objectName(), qualname[child], child

    @classmethod
    def patch(cls, parent):
        if not hasattr(parent, 'widgets'):
            parent.widgets = cls(parent)


# -*- coding: utf-8 -*-
"""
    lantz.ui.featscan
    ~~~~~~~~~~~~~~~~~

    A Feat Scan frontend and Backend. Builds upon Scan.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from ..widgets import WidgetMixin
from ..app import Backend, InstrumentSlot
from .layouts import VerticalUi


class Feat(Backend):
    """A backend to scan a feat for a given instrument."""

    instrument = InstrumentSlot

    #: Name of the scanned feat
    #: :type: str

    def __init__(self, feat_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.feat_name = feat_name

    @property
    def feat_units(self):
        """Units of the scanned feat."""
        target = self.instrument
        feat_name = self.feat_name
        feat = target.feats[feat_name]
        return str(feat.units)


class FeatUi(VerticalUi):
    """A Frontend displaying scan parameters with appropriate units."""

    def connect_backend(self):
        target = self.backend.instrument
        feat_name = self.backend.feat_name

        feat = target.feats[feat_name]

        def _pimp(widget):
            WidgetMixin.wrap(widget)
            widget.bind_feat(feat)

        _pimp(getattr(self.widget, self.backend.feat_name))

        super().connect_backend()





# -*- coding: utf-8 -*-
"""
    lantz.qt.config
    ~~~~~~~~~~~~~~~

    :copyright: 2018 by The Lantz Authors
    :license: BSD, see LICENSE for more details.
"""

import os

from lantz.core.config import register_and_get


# ====================================
# Configuration Values for lantz_sims
# ====================================

# Print the traceback when is not possible to create or bind a qt control.
# valid values: true, false
PRINT_TRACEBACK = register_and_get('qt.print_traceback', 'false') == 'true'

# Qt API wrapper to use in lantz.
# valid values: mock, pyqt, pyqt5, pyqtv1, pyqtdefault, pyside, pyside2
QT_API = register_and_get('qt.api', os.environ.get('QT_API', ''))



# -*- coding: utf-8 -*-
"""
    lantz.qt.widgets
    ~~~~~~~~~~~~~~~~

    PyQt widgets wrapped to work with lantz.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from . import feat, nonnumeric, numeric
from .common import WidgetMixin, ChildrenWidgets
from .initialize import InitializeWindow, InitializeDialog
from .testgui import DriverTestWidget, SetupTestWidget


# -*- coding: utf-8 -*-
"""
    lantz.widgets.dialog_action
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    PyQt Dialog that inspect a function and opens a dialog to ask
    for values for each argument.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import inspect
import json

from ..log import LOGGER
from ..utils.docscrape import NumpyDocString
from ..utils.qt import QtCore, QtGui


class ArgumentsInputDialog(QtGui.QDialog):
    """Dialog to select values for arguments.

    Parameters
    ----------
    func : callable

    parent : PyQt Widget
        parent widget.

    Example
    -------
    You can call it to get the parameters of a function:
    >>> args = ArgumentsInputDialog.get_params(func, parent)
    >>> func(*args)

    Or you can call it to enter the parameters of a function and run it:
    >>> args = ArgumentsInputDialog.run(func, parent)

    """
    def __init__(self, argspec, parent=None, window_title='Function arguments', doc=None):
        super().__init__(parent)

        vlayout = QtGui.QVBoxLayout(self)

        layout = QtGui.QFormLayout()

        widgets = []

        defaults = argspec.defaults if argspec.defaults else ()
        defaults = ('', ) * (len(argspec.args[1:]) - len(defaults)) + defaults

        self.arguments = {}
        for arg, default in zip(argspec.args[1:], defaults):
            wid = QtGui.QLineEdit(self)
            wid.setObjectName(arg)
            wid.setText(json.dumps(default))
            self.arguments[arg] = default

            layout.addRow(arg, wid)
            widgets.append(wid)
            wid.textChanged.connect(self.on_widget_change(wid))
            if doc and arg in doc:
                wid.setToolTip(doc[arg])

        self.widgets = widgets

        buttonBox = QtGui.QDialogButtonBox()
        buttonBox.setOrientation(QtCore.Qt.Horizontal)
        buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Ok)
        buttonBox.setEnabled(True)
        buttonBox.accepted.connect(self.accept)

        vlayout.addLayout(layout)

        label = QtGui.QLabel()
        label.setText('Values are decoded from text using as JSON.')
        vlayout.addWidget(label)

        vlayout.addWidget(buttonBox)

        self.buttonBox = buttonBox
        self.valid = {wid.objectName(): True for wid in self.widgets}

        self.setWindowTitle(window_title)

    def on_widget_change(self, widget):
        name = widget.objectName()
        def validate(value):
            try:
                if value:
                    value = json.loads(value)
                else:
                    value = None
                palette = QtGui.QPalette()
                palette.setColor(widget.backgroundRole(), QtGui.QColor('white'))
                widget.setPalette(palette)
                self.arguments[name] = value
                self.valid[name] = True
            except:
                palette = QtGui.QPalette()
                palette.setColor(widget.backgroundRole(), QtGui.QColor(255, 102, 102))
                widget.setPalette(palette)
                self.valid[name] = False

            self.buttonBox.setEnabled(all(self.valid.values()))

        return validate

    def accept(self):
        super().accept()

    @staticmethod
    def run(func, parent=None):
        """Display a dialog for a function and run
        """

        wrapped = getattr(func, '__wrapped__', func)
        name = wrapped.__name__

        arguments = ArgumentsInputDialog.get_params(func, parent)

        try:
            return func(**arguments)
        except Exception as e:
            LOGGER.exception(e)
            QtGui.QMessageBox.critical(parent, 'Lantz',
                                       'Instrument error while calling {}'.format(name),
                                       QtGui.QMessageBox.Ok,
                                       QtGui.QMessageBox.NoButton)

    @staticmethod
    def get_params(func, parent=None):
        """Creates and display a UnitInputDialog and return new units.
        """

        wrapped = getattr(func, '__wrapped__', func)
        name = wrapped.__name__
        doc = wrapped.__doc__
        argspec = inspect.getargspec(wrapped)

        arguments = {}
        if len(argspec.args) > 1:
            doc = NumpyDocString(doc).get('Parameters', [])
            doc = {k: '\n'.join(v) for (k, _, v) in doc}

            dialog = ArgumentsInputDialog(argspec, parent,
                                          window_title=name + ' arguments',
                                          doc=doc)
            if not dialog.exec_():
                return None

            arguments = dialog.arguments

        return arguments





from ..utils.qt import QtGui, QtCore

QtWidgets = QtGui

from ..app import Frontend


class BorderUi(Frontend):
    """The Loop frontend provides a GUI for the Rich Backend"""

    auto_connect = False

    #: Tuple with the columns
    #: Each element can be:
    #:   - Frontend class: will be connected to the default backend.
    #:   - Front2Back(Frontend class, backend name): will be connect to a specific backend.
    #:   - tuple: will be iterated to obtain the rows.
    north = ()
    west = ()
    south = ()
    east = ()
    center = ()

    def __add(self, layout, parts):
        """Add widgets in parts to layout.

        Parameters
        ----------
        layout :

        parts :


        Returns
        -------

        """
        for part_name in parts:
            if part_name is True:
                layout.addStretch()
                continue

            part = getattr(self, part_name)

            if isinstance(part, Frontend):
                layout.addWidget(part)

            elif isinstance(part, tuple):
                # A tuple found in parts is considered nesting
                if isinstance(layout, self._inner):
                    sublayout = self._outer()
                elif isinstance(layout, self._outer):
                    sublayout = self._inner()
                else:
                    raise ValueError('Unknown parent layout %s' % layout)

                self.__add(sublayout, part)

                layout.setLayout(sublayout)

            else:
                raise ValueError('Only Frontend or tuple are valid values '
                                 'valid for parts not %s (%s)' % (part, type(part)))

        return layout

    def __build_layout(self, content):
        if not content:
            return None
        layout_type, *names = content
        layout = layout_type()
        for name in names:
            if name is True:
                layout.addStretch()
                continue

            part = getattr(self, name)

            if isinstance(part, Frontend):
                layout.addWidget(part)
            else:
                raise ValueError('Only Frontend are valid values '
                                 'valid for parts not %s (%s)' % (part, type(part)))
        return layout

    def setupUi(self):
        super().setupUi()

        center = self.__build_layout(self.center)
        east = self.__build_layout(self.east)
        west = self.__build_layout(self.west)
        south = self.__build_layout(self.south)
        north = self.__build_layout(self.north)
        self.widget = BorderWidget(center, north, south, east, west)
        self.setCentralWidget(self.widget)


class ItemWrapper(object):
    def __init__(self, i, p):
        self.item = i
        self.position = p


class BorderLayout(QtWidgets.QLayout):
    West, North, South, East, Center = range(5)
    MinimumSize, SizeHint = range(2)

    def __init__(self, parent=None, margin=None, spacing=-1):
        super(BorderLayout, self).__init__(parent)

        if margin is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)
        self.list = []

    def __del__(self):
        l = self.takeAt(0)
        while l is not None:
            l = self.takeAt(0)

    def addItem(self, item):
        self.add(item, self.West)

    def addWidget(self, widget, position):
        self.add(QtWidgets.QWidgetItem(widget), position)

    def addLayout(self, layout, position):
        w = QtWidgets.QWidget()
        w.setLayout(layout)
        self.addWidget(w, position)

    def expandingDirections(self):
        return QtGui.Horizontal | QtGui.Vertical

    def hasHeightForWidth(self):
        return False

    def count(self):
        return len(self.list)

    def itemAt(self, index):
        if index < len(self.list):
            return self.list[index].item

        return None

    def minimumSize(self):
        return self.calculateSize(self.MinimumSize)

    def setGeometry(self, rect):
        center = None
        eastWidth = 0
        westWidth = 0
        northHeight = 0
        southHeight = 0
        centerHeight = 0

        super(BorderLayout, self).setGeometry(rect)

        for wrapper in self.list:
            item = wrapper.item
            position = wrapper.position

            if position == self.North:
                item.setGeometry(QtCore.QRect(rect.x(), northHeight,
                        rect.width(), item.sizeHint().height()))

                northHeight += item.geometry().height() + self.spacing()

            elif position == self.South:
                item.setGeometry(QtCore.QRect(item.geometry().x(),
                        item.geometry().y(), rect.width(),
                        item.sizeHint().height()))

                southHeight += item.geometry().height() + self.spacing()

                item.setGeometry(QtCore.QRect(rect.x(),
                        rect.y() + rect.height() - southHeight + self.spacing(),
                        item.geometry().width(), item.geometry().height()))

            elif position == self.Center:
                center = wrapper

        centerHeight = rect.height() - northHeight - southHeight

        for wrapper in self.list:
            item = wrapper.item
            position = wrapper.position

            if position == self.West:
                item.setGeometry(QtCore.QRect(rect.x() + westWidth,
                        northHeight, item.sizeHint().width(), centerHeight))

                westWidth += item.geometry().width() + self.spacing()

            elif position == self.East:
                item.setGeometry(QtCore.QRect(item.geometry().x(),
                        item.geometry().y(), item.sizeHint().width(),
                        centerHeight))

                eastWidth += item.geometry().width() + self.spacing()

                item.setGeometry(QtCore.QRect(rect.x() + rect.width() - eastWidth + self.spacing(),
                        northHeight, item.geometry().width(),
                        item.geometry().height()))

        if center:
            center.item.setGeometry(QtCore.QRect(westWidth, northHeight,
                    rect.width() - eastWidth - westWidth, centerHeight))

    def sizeHint(self):
        return self.calculateSize(self.SizeHint)

    def takeAt(self, index):
        if index >= 0 and index < len(self.list):
            layoutStruct = self.list.pop(index)
            return layoutStruct.item

        return None

    def add(self, item, position):
        self.list.append(ItemWrapper(item, position))

    def calculateSize(self, sizeType):
        totalSize = QtGui.QSize()

        for wrapper in self.list:
            position = wrapper.position
            itemSize = QtGui.QSize()

            if sizeType == self.MinimumSize:
                itemSize = wrapper.item.minimumSize()
            else: # sizeType == self.SizeHint
                itemSize = wrapper.item.sizeHint()

            if position in (self.North, self.South, self.Center):
                totalSize.setHeight(totalSize.height() + itemSize.height())

            if position in (self.West, self.East, self.Center):
                totalSize.setWidth(totalSize.width() + itemSize.width())

        return totalSize


class BorderWidget(QtWidgets.QWidget):

    def __init__(self, center, north, south, east, west):
        super(BorderWidget, self).__init__()

        layout = BorderLayout()
        layout.addLayout(center, BorderLayout.Center)

        # Because BorderLayout doesn't call its super-class addWidget() it
        # doesn't take ownership of the widgets until setLayout() is called.
        # Therefore we keep a local reference to each label to prevent it being
        # garbage collected too soon.

        if north:
            layout.addLayout(north, BorderLayout.North)

        if west:
            layout.addLayout(west, BorderLayout.West)

        if east:
            layout.addLayout(east, BorderLayout.East)

        if south:
            layout.addLayout(south, BorderLayout.South)

        self.setLayout(layout)




# -*- coding: utf-8 -*-
"""
    lantz.widgets.test
    ~~~~~~~~~~~~~~~~~~

    Numeric widgets.
    - QDoubleSpinBox
    - CheckBox
    - QLineEdit
    - QSlider
    - QDial
    - QProgressBar
    - QScrollBar
    - QLCDNumber

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from lantz.core import Driver

from ..utils.qt import QtCore, QtGui

from ..log import LOGGER
from ..config import PRINT_TRACEBACK
from ..utils.qt import QtGui
from .feat import LabeledFeatWidget
from .dialog_action import ArgumentsInputDialog


class DriverTestWidget(QtGui.QWidget):
    """Widget that is automatically filled to control all Feats of a given driver.

    Parameters
    ----------
    parent : PyQt Widget
        parent widget.
    target : Lantz
        driver object to map.

    """

    def __init__(self, parent, target):
        super().__init__(parent)
        self._lantz_target = target

        layout = QtGui.QVBoxLayout(self)

        label = QtGui.QLabel()
        label.setText('%s (%s)' % (target.name, target.__class__.__qualname__))
        layout.addWidget(label)

        recall = QtGui.QPushButton()
        recall.setText('Refresh')
        recall.clicked.connect(lambda x: target.refresh())

        update = QtGui.QPushButton()
        update.setText('Update')
        update.clicked.connect(lambda x: target.update(self.widgets_values_as_dict()))

        auto = QtGui.QCheckBox()
        auto.setText('Update on change')
        auto.setChecked(True)
        auto.stateChanged.connect(self.update_on_change)

        hlayout = QtGui.QHBoxLayout()
        hlayout.addWidget(recall)
        hlayout.addWidget(update)
        hlayout.addWidget(auto)

        layout.addLayout(hlayout)

        self.writable_widgets = []
        self.widgets = []

        # Feat
        for feat_name, feat in sorted(target.feats.items()):
            try:
                feat_widget = LabeledFeatWidget(self, target, feat)

                self.widgets.append(feat_widget)
                if feat_widget.writable:
                    self.writable_widgets.append(feat_widget)

                layout.addWidget(feat_widget)
            except Exception as ex:
                LOGGER.debug('Could not create control for {}: {}'.format(feat_name, ex))
                if PRINT_TRACEBACK:
                    import traceback
                    traceback.print_exc()

        # DictFeats
        for feat_name, feat in sorted(target.dictfeats.items()):
            try:
                feat_widget = LabeledFeatWidget(self, target, feat)

                self.widgets.append(feat_widget)
                if feat_widget.writable:
                    self.writable_widgets.append(feat_widget)

                layout.addWidget(feat_widget)
            except Exception as ex:
                LOGGER.debug('Could not create control for {}: {}'.format(feat_name, ex))
                if PRINT_TRACEBACK:
                    import traceback
                    traceback.print_exc()

        # Actions
        line = QtGui.QFrame(self)
        #self.line.setGeometry(QtCore.QRect(110, 80, 351, 31))
        line.setFrameShape(QtGui.QFrame.HLine)
        line.setFrameShadow(QtGui.QFrame.Sunken)
        layout.addWidget(line)

        actions_label = QtGui.QLabel(self)
        actions_label.setText('Actions:')
        actions_label.setFixedWidth(120)

        self.actions_combo = QtGui.QComboBox(self)
        actions = [n for n in target.actions.keys() if n not in set(Driver._lantz_actions.keys())]
        self.actions_combo.addItems(actions)

        actions_button = QtGui.QPushButton(self)
        actions_button.setFixedWidth(60)
        actions_button.setText('Run')
        actions_button.clicked.connect(self.on_run_clicked)

        alayout = QtGui.QHBoxLayout()
        alayout.addWidget(actions_label)
        alayout.addWidget(self.actions_combo)
        alayout.addWidget(actions_button)

        layout.addLayout(alayout)

        self.statusBar = QtGui.QLabel()
        self.statusBar.setText('Ready ...')
        layout.addWidget(self.statusBar)

    @QtCore.Slot()
    def on_run_clicked(self):
        func = getattr(self._lantz_target, self.actions_combo.currentText())
        args = ArgumentsInputDialog.get_params(func, self)
        self.statusBar.setText('Running ...')
        try:
            out = func(**args)
            out = 'Return value: %s' % out
        except Exception as e:
            out = 'Error: %s' % e

        self.statusBar.setText(out)

    def update_on_change(self, new_state):
        """Set the 'update_on_change' flag to new_state in each writable widget
        within this widget. If True, the driver will be updated after each change.
        """

        for widget in self.writable_widgets:
            widget._widget._update_on_change = new_state

    def widgets_values_as_dict(self):

        return {widget.feat.name: widget._widget.value()
                for widget in self.writable_widgets}

    @property
    def lantz_target(self):
        """Driver connected to this widget."""
        return self._lantz_target

    @lantz_target.setter
    def lantz_target(self, driver):
        self._lantz_target = driver
        for widget in self.widgets:
            widget.lantz_target = driver


class SetupTestWidget(QtGui.QWidget):
    """Widget to control multiple drivers.

    Parameters
    ----------
    parent :
        parent widget.
    targets :
        iterable of driver object to map.
    """

    def __init__(self, parent, targets):
        super().__init__(parent)

        layout = QtGui.QHBoxLayout(self)

        tab_widget = QtGui.QTabWidget(self)
        tab_widget.setTabsClosable(False)
        for target in targets:
            widget = DriverTestWidget(parent, target)
            tab_widget.addTab(widget, target.name)

        layout.addWidget(tab_widget)



# -*- coding: utf-8 -*-
"""
    lantz.qt.app
    ~~~~~~~~~~~~

    Wraps a Lantz Driver in a PyQt object connecting `_changed` signals.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .utils.qt import QtCore, SuperQObject


class QDriver(QtCore.QObject):
    pass


def wrap_driver(wrapped_obj, parent_qobj=None):

    SUFFIX = '_changed'
    PYSUFFIX = '_py_changed'

    class QObj(QDriver):

        def __init__(self, obj, parent):
            super().__init__(parent)
            self.wrapped_obj = obj

        # Qt Signals need to be added to the class before it is created.
        # We loop through all members of the class and add a changed event
        # for each Feat/DictFeat.

        # We do the same thing for all base classes which are not derived from
        # Driver (checking for the attribute _lantz_feats) to enable base clases
        # for drivers that do not derive from Driver.

        def __getattr__(self, item):
            if item in self.__dict__:
                return getattr(self, item)

            return getattr(self.wrapped_obj, item)

        for feat_name in wrapped_obj.feats.keys():
            #locals()[feat_name + PYSUFFIX] = getattr(wrapped_obj, feat_name + SUFFIX)
            locals()[feat_name + SUFFIX] = QtCore.pyqtSignal(object, object)

        for feat_name in wrapped_obj.dictfeats.keys():
            #locals()[feat_name + PYSUFFIX] = getattr(wrapped_obj, feat_name + SUFFIX)
            locals()[feat_name + SUFFIX] = QtCore.pyqtSignal(object, object, object)

    obj = QObj(wrapped_obj, parent_qobj)

    return obj


def wrap_driver_cls(wrapped_cls):

    SUFFIX = '_changed'
    PYSUFFIX = '_py_changed'

    class WrappedCLS(wrapped_cls, SuperQObject):

        # Qt Signals need to be added to the class before it is created.
        # We loop through all members of the class and add a changed event
        # for each Feat/DictFeat.

        for feat_name in wrapped_cls._lantz_feats.keys():
            #locals()[feat_name + PYSUFFIX] = getattr(wrapped_obj, feat_name + SUFFIX)
            locals()[feat_name + SUFFIX] = QtCore.pyqtSignal(object, object)

        for feat_name in wrapped_cls._lantz_dictfeats.keys():
            #locals()[feat_name + PYSUFFIX] = getattr(wrapped_obj, feat_name + SUFFIX)
            locals()[feat_name + SUFFIX] = QtCore.pyqtSignal(object, object, object)

    WrappedCLS.__name__ = 'Qt' + wrapped_cls.__name__

    return WrappedCLS



# -*- coding: utf-8 -*-
"""
    lantz.qt.blocks
    ~~~~~~~~~~~~~~~

    Building blocks for rich applications.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .loop import Loop, LoopUi
from .scan import Scan, ScanUi
from .feat import Feat, FeatUi
from .featscan import FeatScan, FeatScanUi
from .chart import ChartUi
from .layouts import HorizonalUi, VerticalUi, ToolbarLeftRightUi
from .wlayout import BorderUi


# -*- coding: utf-8 -*-
"""
    lantz.utils.qt_loaders
    ~~~~~~~~~~~~~~~~~~~~~~

    This module contains factory functions that attempt
    to return Qt submodules from the various python Qt bindings.

    It also protects against double-importing Qt with different
    bindings, which is unstable and likely to crash

    This is used primarily by qt and qt_for_kernel, and shouldn't
    be accessed directly from the outside

    Copied with modifications from the Jupyter Project.
    https://raw.githubusercontent.com/jupyter/qtconsole/master/qtconsole/qt_loaders.py

    :copyright: IPython
    :license: BSD, see the Jupyter Project for more details.
"""

QT_API_MOCK = 'mock'


def import_qtmock():
    """ """
    # The new-style string API (version=2) automatically
    # converts QStrings to Unicode Python strings. Also, automatically unpacks
    # QVariants to their underlying objects.

    from unittest.mock import MagicMock

    class Mock(MagicMock):

        @classmethod
        def __getattr__(cls, name):
            m = Mock()
            m.__name__ = name
            return m

        QObject = object

    QtGui, QtCore, QtSvg = Mock(), Mock(), Mock()

    return QtCore, QtGui, QtSvg, 'mock'


#### Start Copied


"""
This module contains factory functions that attempt
to return Qt submodules from the various python Qt bindings.

It also protects against double-importing Qt with different
bindings, which is unstable and likely to crash

This is used primarily by qt and qt_for_kernel, and shouldn't
be accessed directly from the outside
"""
import sys
import types
from functools import partial

from distutils.version import LooseVersion

def check_version(a, b):
    """compare versions"""
    try:
        return LooseVersion(a) >= LooseVersion(b)
    except TypeError:
        # assume unparseable versions are latest dev
        return True

# Available APIs.
QT_API_PYQT = 'pyqt'
QT_API_PYQT5 = 'pyqt5'
QT_API_PYQTv1 = 'pyqtv1'
QT_API_PYQT_DEFAULT = 'pyqtdefault' # don't set SIP explicitly
QT_API_PYSIDE = 'pyside'
QT_API_PYSIDE2 = 'pyside2'

api_to_module = {QT_API_PYSIDE2: 'PySide2',
                 QT_API_PYSIDE: 'PySide',
                 QT_API_PYQT: 'PyQt4',
                 QT_API_PYQTv1: 'PyQt4',
                 QT_API_PYQT5: 'PyQt5',
                 QT_API_PYQT_DEFAULT: 'PyQt4',
                 QT_API_MOCK: 'mock'
                }

class ImportDenier(object):
    """Import Hook that will guard against bad Qt imports
    once committed to a specific binding
    """

    def __init__(self):
        self.__forbidden = set()

    def forbid(self, module_name):
        sys.modules.pop(module_name, None)
        self.__forbidden.add(module_name)

    def find_module(self, fullname, path=None):
        if path:
            return
        if fullname in self.__forbidden:
            return self

    def load_module(self, fullname):
        raise ImportError("""
    Importing %s disabled by QtConsole, which has
    already imported an incompatible Qt Binding: %s
    """ % (fullname, loaded_api()))

ID = ImportDenier()
sys.meta_path.append(ID)


def commit_api(api):
    """Commit to a particular API, and trigger ImportErrors on subsequent
       dangerous imports"""

    if api == QT_API_PYSIDE2:
        ID.forbid('PySide')
        ID.forbid('PyQt4')
        ID.forbid('PyQt5')
    if api == QT_API_PYSIDE:
        ID.forbid('PySide2')
        ID.forbid('PyQt4')
        ID.forbid('PyQt5')
    elif api == QT_API_PYQT5:
        ID.forbid('PySide2')
        ID.forbid('PySide')
        ID.forbid('PyQt4')
    else:   # There are three other possibilities, all representing PyQt4
        ID.forbid('PySide2')
        ID.forbid('PyQt5')
        ID.forbid('PySide')


def loaded_api():
    """Return which API is loaded, if any

    If this returns anything besides None,
    importing any other Qt binding is unsafe.

    Returns
    -------
    None, 'pyside2', 'pyside', 'pyqt', or 'pyqtv1'
    """
    if 'PyQt4.QtCore' in sys.modules:
        if qtapi_version() == 2:
            return QT_API_PYQT
        else:
            return QT_API_PYQTv1
    elif 'PySide.QtCore' in sys.modules:
        return QT_API_PYSIDE
    elif 'PySide2.QtCore' in sys.modules:
        return QT_API_PYSIDE2
    elif 'PyQt5.QtCore' in sys.modules:
        return QT_API_PYQT5
    return None


def has_binding(api):
    """Safely check for PyQt4/5, PySide or PySide2, without importing
       submodules

    Supports Python <= 3.3

       Parameters
       ----------
       api : str [ 'pyqtv1' | 'pyqt' | 'pyqt5' | 'pyside' | 'pyside2' | 'pyqtdefault']
            Which module to check for

       Returns
       -------
       True if the relevant module appears to be importable
    """
    # we can't import an incomplete pyside and pyqt4
    # this will cause a crash in sip (#1431)
    # check for complete presence before importing
    module_name = api_to_module[api]

    import imp
    try:
        #importing top level PyQt4/PySide module is ok...
        mod = __import__(module_name)
        #...importing submodules is not
        imp.find_module('QtCore', list(mod.__path__))
        imp.find_module('QtGui', list(mod.__path__))
        imp.find_module('QtSvg', list(mod.__path__))
        if api == QT_API_PYQT5 or api == QT_API_PYSIDE2:
            # QT5 requires QtWidgets too
            imp.find_module('QtWidgets', list(mod.__path__))

        #we can also safely check PySide version
        if api == QT_API_PYSIDE:
            return check_version(mod.__version__, '1.0.3')
        else:
            return True
    except ImportError:
        return False

def has_binding_new(api):
    """Safely check for PyQt4/5 or PySide, without importing submodules

    Supports Python >= 3.4

        Parameters
        ----------
        api : str [ 'pyqtv1' | 'pyqt' | 'pyqt5' | 'pyside' | 'pyqtdefault']
             Which module to check for

        Returns
        -------
        True if the relevant module appears to be importable
     """
    module_name = api_to_module[api]
    from importlib.util import find_spec

    required = ['QtCore', 'QtGui', 'QtSvg']
    if api in (QT_API_PYQT5, QT_API_PYSIDE2):
        # QT5 requires QtWidgets too
        required.append('QtWidgets')

    for submod in required:
        try:
            spec = find_spec('%s.%s' % (module_name, submod))
        except ImportError:
            # Package (e.g. PyQt5) not found
            return False
        else:
            if spec is None:
                # Submodule (e.g. PyQt5.QtCore) not found
                return False

    if api == QT_API_PYSIDE:
        # We can also safely check PySide version
        import PySide
        return check_version(PySide.__version__, '1.0.3')

    return True

if sys.version_info >= (3, 4):
    has_binding = has_binding_new

def qtapi_version():
    """Return which QString API has been set, if any

    Returns
    -------
    The QString API version (1 or 2), or None if not set
    """
    try:
        import sip
    except ImportError:
        return
    try:
        return sip.getapi('QString')
    except ValueError:
        return


def can_import(api):
    """Safely query whether an API is importable, without importing it"""
    if not has_binding(api):
        return False

    current = loaded_api()
    if api == QT_API_PYQT_DEFAULT:
        return current in [QT_API_PYQT, QT_API_PYQTv1, None]
    else:
        return current in [api, None]


def import_pyqt4(version=2):
    """
    Import PyQt4

    Parameters
    ----------
    version : 1, 2, or None
      Which QString/QVariant API to use. Set to None to use the system
      default

    ImportErrors rasied within this function are non-recoverable
    """
    # The new-style string API (version=2) automatically
    # converts QStrings to Unicode Python strings. Also, automatically unpacks
    # QVariants to their underlying objects.
    import sip

    if version is not None:
        sip.setapi('QString', version)
        sip.setapi('QVariant', version)

    from PyQt4 import QtGui, QtCore, QtSvg

    if not check_version(QtCore.PYQT_VERSION_STR, '4.7'):
        raise ImportError("QtConsole requires PyQt4 >= 4.7, found %s" %
                          QtCore.PYQT_VERSION_STR)

    # Alias PyQt-specific functions for PySide compatibility.
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot

    # query for the API version (in case version == None)
    version = sip.getapi('QString')
    api = QT_API_PYQTv1 if version == 1 else QT_API_PYQT
    return QtCore, QtGui, QtSvg, api


def import_pyqt5():
    """
    Import PyQt5

    ImportErrors rasied within this function are non-recoverable
    """
    import sip

    from PyQt5 import QtCore, QtSvg, QtWidgets, QtGui, QtPrintSupport

    # Alias PyQt-specific functions for PySide compatibility.
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot

    # Join QtGui and QtWidgets for Qt4 compatibility.
    QtGuiCompat = types.ModuleType('QtGuiCompat')
    QtGuiCompat.__dict__.update(QtGui.__dict__)
    QtGuiCompat.__dict__.update(QtWidgets.__dict__)
    QtGuiCompat.__dict__.update(QtPrintSupport.__dict__)

    api = QT_API_PYQT5
    return QtCore, QtGuiCompat, QtSvg, api


def import_pyside():
    """
    Import PySide

    ImportErrors raised within this function are non-recoverable
    """
    from PySide import QtGui, QtCore, QtSvg
    return QtCore, QtGui, QtSvg, QT_API_PYSIDE


def import_pyside2():
    """
    Import PySide2

    ImportErrors raised within this function are non-recoverable
    """
    from PySide2 import QtGui, QtCore, QtSvg, QtWidgets, QtPrintSupport

    # Join QtGui and QtWidgets for Qt4 compatibility.
    QtGuiCompat = types.ModuleType('QtGuiCompat')
    QtGuiCompat.__dict__.update(QtGui.__dict__)
    QtGuiCompat.__dict__.update(QtWidgets.__dict__)
    QtGuiCompat.__dict__.update(QtPrintSupport.__dict__)

    return QtCore, QtGuiCompat, QtSvg, QT_API_PYSIDE2


def load_qt(api_options):
    """
    Import Qt, given a preference list of permissible bindings.

    It is safe to call this function multiple times.  If a Qt binding
    has already been loaded, it will be returned regardless of the
    preference list.

    Parameters
    ----------
    api_options: List of strings
        The order of APIs to try. Valid items are 'pyside2', 'pyside', 'pyqt',
        'pyqt5', 'pyqtv1' and 'pyqtdefault'

    Returns
    -------

    A tuple of QtCore, QtGui, QtSvg, QT_API
    The first three are the Qt modules. The last is the string
    indicating which module was loaded.

    Raises
    ------
    ImportError, if it isn't possible to import any requested bindings
    (because they aren't installed).
    """

    loaders = {QT_API_PYSIDE2: import_pyside2,
               QT_API_PYSIDE: import_pyside,
               QT_API_PYQT: import_pyqt4,
               QT_API_PYQT5: import_pyqt5,
               QT_API_PYQTv1: partial(import_pyqt4, version=1),
               QT_API_PYQT_DEFAULT: partial(import_pyqt4, version=None),
               QT_API_MOCK: import_qtmock,
               }

    loaded = loaded_api()
    if loaded:
        api_options = [loaded]

    for api in api_options:

        if api not in loaders:
            raise RuntimeError(
                "Invalid Qt API %r, valid values are: %s" %
                (api, ", ".join(["%r" % k for k in loaders.keys()])))

        if not can_import(api):
            continue

        #cannot safely recover from an ImportError during this
        result = loaders[api]()
        api = result[-1]  # changed if api = QT_API_PYQT_DEFAULT
        commit_api(api)
        return result
    else:
        raise ImportError("""
    Could not load requested Qt binding. Please ensure that
    PyQt4 >= 4.7, PyQt5, PySide >= 1.0.3 or PySide2 is available,
    and only one is imported per session.

    Currently-imported Qt library:   %r
    PyQt4 installed:                 %s
    PyQt5 installed:                 %s
    PySide >= 1.0.3 installed:       %s
    PySide2 installed:               %s
    Tried to load:                   %r
    """ % (loaded,
           has_binding(QT_API_PYQT),
           has_binding(QT_API_PYQT5),
           has_binding(QT_API_PYSIDE),
           has_binding(QT_API_PYSIDE2),
           api_options))

def load_qtest():
    """Load QTest module from the loaded QT API."""
    api = loaded_api()
    if api == QT_API_PYQT5:
        from PyQt5.QtTest import QTest
    elif api == QT_API_PYSIDE:
        from PySide.QtTest import QTest
    elif api == QT_API_PYSIDE2:
        from PySide2.QtTest import QTest
    else:
        from PyQt4.QtTest import QTest
    return QTest



# -*- coding: utf-8 -*-
"""
    lantz.widgets.nonnumeric
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Non-umeric widgets.
    - QComboBox
    - QCheckBox
    - QQLineEdit

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import json

from lantz.core.helpers import MISSING, UNSET

from ..utils.qt import QtGui
from .common import WidgetMixin, register_wrapper


@register_wrapper
class QComboBoxMixin(WidgetMixin):

    _WRAPPED = (QtGui.QComboBox, )

    @classmethod
    def _wrap(cls, widget):
        super()._wrap(widget)
        widget.valueChanged = widget.currentIndexChanged

    def value(self):
        return self.currentText()

    def setValue(self, value):
        if value is MISSING or value is UNSET:
            font = QtGui.QFont()
            font.setItalic(True)
            self.setFont(font)
            return
        self.setCurrentIndex(self.__values.index(value))

    def setReadOnly(self, value):
        self.setEnabled(not value)

    def bind_feat(self, feat):
        super().bind_feat(feat)
        if isinstance(self._feat.values, dict):
            self.__values = list(self._feat.values.keys())
        else:
            self.__values = list(self.__values)
        self.clear()
        self.addItems([str(value) for value in self.__values])


@register_wrapper
class QCheckBoxMixin(WidgetMixin):

    _WRAPPED = (QtGui.QCheckBox, )

    @classmethod
    def _wrap(cls, widget):
        super()._wrap(widget)
        widget.valueChanged = widget.stateChanged

    def setReadOnly(self, value):
        self.setCheckable(not value)

    def value(self):
        return self.isChecked()

    def setValue(self, value):
        if value is MISSING or value is UNSET:
            return
        self.setChecked(value)


@register_wrapper
class QLineEditMixin(WidgetMixin):

    _WRAPPED = (QtGui.QLineEdit, )

    @classmethod
    def _wrap(cls, widget):
        super()._wrap(widget)
        widget.valueChanged = widget.textChanged

    def value(self):
        return self.text()

    def setValue(self, value):
        if value is MISSING or value is UNSET:
            return
        if isinstance(value, dict) and not self.writable:
            return self.setText(json.dumps(value))
        return self.setText(value)





# Adapted from the qtpy project
# See https://raw.githubusercontent.com/spyder-ide/qtpy/master/qtpy/uic.py

from . import qt_loaders


def build_loadUi(QT_API):

    if QT_API == qt_loaders.QT_API_PYQT5:

        from PyQt5.uic import loadUi, loadUiType

        return loadUi, loadUiType

    elif QT_API == qt_loaders.QT_API_PYQT:

        from PyQt4.uic import loadUi, loadUiType

        return loadUi, loadUiType

    else:

        # In PySide, loadUi does not exist, so we define it using QUiLoader, and
        # then make sure we expose that function. This is adapted from qt-helpers
        # which was released under a 3-clause BSD license:
        # qt-helpers - a common front-end to various Qt modules
        #
        # Copyright (c) 2015, Chris Beaumont and Thomas Robitaille
        #
        # All rights reserved.
        #
        # Redistribution and use in source and binary forms, with or without
        # modification, are permitted provided that the following conditions are
        # met:
        #
        #  * Redistributions of source code must retain the above copyright
        #    notice, this list of conditions and the following disclaimer.
        #  * Redistributions in binary form must reproduce the above copyright
        #    notice, this list of conditions and the following disclaimer in the
        #    documentation and/or other materials provided with the
        #    distribution.
        #  * Neither the name of the Glue project nor the names of its contributors
        #    may be used to endorse or promote products derived from this software
        #    without specific prior written permission.
        #
        # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
        # IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
        # THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
        # PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
        # CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
        # EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
        # PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
        # PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
        # LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
        # NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
        #
        # Which itself was based on the solution at
        #
        # https://gist.github.com/cpbotha/1b42a20c8f3eb9bb7cb8
        #
        # which was released under the MIT license:
        #
        # Copyright (c) 2011 Sebastian Wiesner <lunaryorn@gmail.com>
        # Modifications by Charl Botha <cpbotha@vxlabs.com>
        #
        # Permission is hereby granted, free of charge, to any person obtaining a
        # copy of this software and associated documentation files (the "Software"),
        # to deal in the Software without restriction, including without limitation
        # the rights to use, copy, modify, merge, publish, distribute, sublicense,
        # and/or sell copies of the Software, and to permit persons to whom the
        # Software is furnished to do so, subject to the following conditions:
        #
        # The above copyright notice and this permission notice shall be included in
        # all copies or substantial portions of the Software.
        #
        # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
        # THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
        # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
        # DEALINGS IN THE SOFTWARE.

        if QT_API == qt_loaders.QT_API_PYSIDE:
            from PySide.QtCore import QMetaObject
            from PySide.QtUiTools import QUiLoader
        elif QT_API == qt_loaders.QT_API_PYSIDE2:
            from PySide2.QtCore import QMetaObject
            from PySide2.QtUiTools import QUiLoader

        class UiLoader(QUiLoader):
            """
            Subclass of :class:`~PySide.QtUiTools.QUiLoader` to create the user
            interface in a base instance.

            Unlike :class:`~PySide.QtUiTools.QUiLoader` itself this class does not
            create a new instance of the top-level widget, but creates the user
            interface in an existing instance of the top-level class if needed.

            This mimics the behaviour of :func:`PyQt4.uic.loadUi`.
            """

            def __init__(self, baseinstance, customWidgets=None):
                """
                Create a loader for the given ``baseinstance``.

                The user interface is created in ``baseinstance``, which must be an
                instance of the top-level class in the user interface to load, or a
                subclass thereof.

                ``customWidgets`` is a dictionary mapping from class name to class
                object for custom widgets. Usually, this should be done by calling
                registerCustomWidget on the QUiLoader, but with PySide 1.1.2 on
                Ubuntu 12.04 x86_64 this causes a segfault.

                ``parent`` is the parent object of this loader.
                """

                QUiLoader.__init__(self, baseinstance)

                self.baseinstance = baseinstance

                if customWidgets is None:
                    self.customWidgets = {}
                else:
                    self.customWidgets = customWidgets

            def createWidget(self, class_name, parent=None, name=''):
                """
                Function that is called for each widget defined in ui file,
                overridden here to populate baseinstance instead.
                """

                if parent is None and self.baseinstance:
                    # supposed to create the top-level widget, return the base
                    # instance instead
                    return self.baseinstance

                else:

                    # For some reason, Line is not in the list of available
                    # widgets, but works fine, so we have to special case it here.
                    if class_name in self.availableWidgets() or class_name == 'Line':
                        # create a new widget for child widgets
                        widget = QUiLoader.createWidget(self, class_name, parent, name)

                    else:
                        # If not in the list of availableWidgets, must be a custom
                        # widget. This will raise KeyError if the user has not
                        # supplied the relevant class_name in the dictionary or if
                        # customWidgets is empty.
                        try:
                            widget = self.customWidgets[class_name](parent)
                        except KeyError:
                            raise Exception('No custom widget ' + class_name + ' '
                                            'found in customWidgets')

                    if self.baseinstance:
                        # set an attribute for the new child widget on the base
                        # instance, just like PyQt4.uic.loadUi does.
                        setattr(self.baseinstance, name, widget)

                    return widget

        def _get_custom_widgets(ui_file):
            """
            This function is used to parse a ui file and look for the <customwidgets>
            section, then automatically load all the custom widget classes.
            """

            import sys
            import importlib
            from xml.etree.ElementTree import ElementTree

            # Parse the UI file
            etree = ElementTree()
            ui = etree.parse(ui_file)

            # Get the customwidgets section
            custom_widgets = ui.find('customwidgets')

            if custom_widgets is None:
                return {}

            custom_widget_classes = {}

            for custom_widget in custom_widgets.getchildren():

                cw_class = custom_widget.find('class').text
                cw_header = custom_widget.find('header').text

                module = importlib.import_module(cw_header)

                custom_widget_classes[cw_class] = getattr(module, cw_class)

            return custom_widget_classes

        def loadUi(uifile, baseinstance=None, workingDirectory=None):
            """
            Dynamically load a user interface from the given ``uifile``.

            ``uifile`` is a string containing a file name of the UI file to load.

            If ``baseinstance`` is ``None``, the a new instance of the top-level
            widget will be created. Otherwise, the user interface is created within
            the given ``baseinstance``. In this case ``baseinstance`` must be an
            instance of the top-level widget class in the UI file to load, or a
            subclass thereof. In other words, if you've created a ``QMainWindow``
            interface in the designer, ``baseinstance`` must be a ``QMainWindow``
            or a subclass thereof, too. You cannot load a ``QMainWindow`` UI file
            with a plain :class:`~PySide.QtGui.QWidget` as ``baseinstance``.

            :method:`~PySide.QtCore.QMetaObject.connectSlotsByName()` is called on
            the created user interface, so you can implemented your slots according
            to its conventions in your widget class.

            Return ``baseinstance``, if ``baseinstance`` is not ``None``. Otherwise
            return the newly created instance of the user interface.
            """

            # We parse the UI file and import any required custom widgets
            customWidgets = _get_custom_widgets(uifile)

            loader = UiLoader(baseinstance, customWidgets)

            if workingDirectory is not None:
                loader.setWorkingDirectory(workingDirectory)

            widget = loader.load(uifile)
            QMetaObject.connectSlotsByName(widget)
            return widget

        def loadUiType(*args, **kwargs):
            raise Exception('QT_API=%s does not implement loadUiType yet.' % QT_API)

        return loadUi, loadUiType


# -*- coding: utf-8 -*-
"""
    lantz.widgets.feat
    ~~~~~~~~~~~~~~~~~~

    Widgets to wrap Feat and DictFeat

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import inspect

from lantz.core.feat import DictFeat, DictFeatProxy
from lantz.core.helpers import UNSET, MISSING

from ..utils.qt import QtCore, QtGui
from .common import WidgetMixin


class FeatWidget(object):
    """Widget to show a Feat.

    Parameters
    ----------
    parent :
        parent widget.
    target :
        driver object to connect.
    feat :
        DictFeat to connect.
    """


    def __new__(cls, parent, target, feat):
        widget = WidgetMixin.from_feat(feat, parent)
        widget.bind_feat(feat)
        widget.lanz_target = target
        return widget


class DictFeatWidget(QtGui.QWidget):
    """Widget to show a DictFeat.

    Parameters
    ----------
    parent :
        parent widget.
    target :
        driver object to connect.
    feat :
        DictFeat to connect.
    """

    def __init__(self, parent, target, feat):
        super().__init__(parent)

        if isinstance(feat, DictFeatProxy):
            self._feat = feat.proxied
        elif isinstance(feat, DictFeat):
            self._feat = feat
        else:
            raise ValueError('feat argument must be an instance of DictFeat or DictFeatProxy, not %s' % feat.__class__)

        layout = QtGui.QHBoxLayout(self)

        if feat.keys:
            self._key_widget = QtGui.QComboBox()
            if isinstance(feat.keys, dict):
                self._keys = list(feat.keys.keys())
            else:
                self._keys = list(feat.keys)

            self._key_widget.addItems([str(key) for key in self._keys])
            self._key_widget.currentIndexChanged.connect(self._combobox_changed)
        else:
            self._key_widget = QtGui.QLineEdit()
            self._key_widget.textChanged.connect(self._lineedit_changed)

        layout.addWidget(self._key_widget)

        self._value_widget = WidgetMixin.from_feat(feat)
        self._value_widget.bind_feat(feat)
        self._value_widget.feat_key = self._keys[0] if self._keys else MISSING
        self._value_widget.lantz_target = target

        layout.addWidget(self._value_widget)

        self.widgets = (self._key_widget, self._value_widget)

    @QtCore.Slot(int)
    def _combobox_changed(self, value):
        self._value_widget.feat_key = self._keys[self._key_widget.currentIndex()]

    @QtCore.Slot(str)
    def _lineedit_changed(self, value):
        self._value_widget.feat_key = self._key_widget.text()

    def value(self):
        """Get widget value."""
        return self._value_widget.value()

    def setValue(self, value):
        """Set widget value."""
        if value is MISSING:
            return
        self._value_widget.setValue(value)

    def setReadOnly(self, value):
        """Set read only."""
        self._value_widget.setReadOnly(value)

    @property
    def lantz_target(self):
        """Driver connected to this widget."""
        return self._value_widget._lantz_target

    @lantz_target.setter
    def lantz_target(self, driver):
        self._value_widget._lantz_target = driver

    @property
    def readable(self):
        """If the Feat associated with the widget can be read (get)."""
        return self._value_widget.readable

    @property
    def writable(self):
        """If the Feat associated with the widget can be written (set)."""
        return self._value_widget.writable

    def value_from_feat(self):
        return self._value_widget.value_from_feat()

    def value_to_feat(self):
        return self._value_widget.value_to_feat()

class LabeledFeatWidget(QtGui.QWidget):
    """Widget containing a label, a control, and a get a set button.

    Parameters
    ----------
    parent :
        parent widget.
    target :
        driver object to connect.
    feat :
        Feat to connect.
    """

    def __init__(self, parent, target, feat):
        super().__init__(parent)

        layout = QtGui.QHBoxLayout(self)

        self._label = QtGui.QLabel()
        self._label.setText(feat.name)
        self._label.setFixedWidth(120)
        self._label.setToolTip(inspect.getdoc(feat))
        layout.addWidget(self._label)

        if isinstance(feat, DictFeatProxy):
            self._widget = DictFeatWidget(parent, target, feat)
        else:
            self._widget = WidgetMixin.from_feat(feat)
            self._widget.bind_feat(feat)
            self._widget.lantz_target = target

        layout.addWidget(self._widget)

        self._get = QtGui.QPushButton()
        self._get.setText('get')
        self._get.setEnabled(self._widget.readable)
        self._get.setFixedWidth(60)
        layout.addWidget(self._get)

        self._set = QtGui.QPushButton()
        self._set.setText('set')
        self._set.setEnabled(self._widget.writable)
        self._set.setFixedWidth(60)
        layout.addWidget(self._set)

        self._get.clicked.connect(self.on_get_clicked)
        self._set.clicked.connect(self.on_set_clicked)
        self._widget._update_on_change = self._widget.writable

        self.widgets = (self._label, self._widget, self._get, self._set)

    @property
    def feat(self):
        return self._widget.feat

    @property
    def label_width(self):
        """Width of the label"""
        return self._label.width

    @label_width.setter
    def label_width(self, value):
        self._label.setFixedWidth(value)

    @property
    def lantz_target(self):
        """Driver connected to this widget."""
        return self._widget._lantz_target

    @lantz_target.setter
    def lantz_target(self, driver):
        self._widget._lantz_target = driver

    @QtCore.Slot()
    def on_get_clicked(self):
        self._widget.value_from_feat()

    @QtCore.Slot()
    def on_set_clicked(self):
        font = QtGui.QFont()
        font.setItalic(False)
        self._widget.setFont(font)
        self._widget.value_to_feat()

    @property
    def readable(self):
        """If the Feat associated with the widget can be read (get)."""
        return self._widget.readable

    @property
    def writable(self):
        """If the Feat associated with the widget can be written (set)."""
        return self._widget.writable





from ..utils.qt import QtGui


class InitializeWindow(QtGui.QWidget):
    """This windows receives the list of drivers to initialize
    and their dependency.
    """

    def __init__(self, drivers=None, dependencies=None, parent=None):
        super(InitializeWindow, self).__init__(parent)

        factory = QtGui.QItemEditorFactory()
        QtGui.QItemEditorFactory.setDefaultFactory(factory)

        self.drivers = drivers
        self.dependencies = dependencies

        self.createGUI()

    @classmethod
    def from_flock(cls, flock):
        return cls(list(flock.values()), flock.dependencies)

    def initialize(self):
        from ..connect import connect_initialize
        self.thread = connect_initialize(self.widget, self.drivers,
                                         dependencies=self.dependencies,
                                         concurrent=True)

    def createGUI(self):

        # Demonstrate the supported widgets.
        # Uncomment to try others.
        self.widget = self._createGUITable()
        #self.widget = self._createGUILine()
        #self.widget = self._createGUIText()

        button = QtGui.QPushButton()
        button.setText('Initialize')
        button.setEnabled(True)
        button.clicked.connect(self.initialize)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(button)
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.setWindowTitle("Driver initialization")

    def _createGUIText(self):

        text = QtGui.QTextEdit()
        return text

    def _createGUILine(self):

        text = QtGui.QLineEdit()
        return text

    def _createGUITable(self):

        table = QtGui.QTableWidget(0, 3)
        table.setHorizontalHeaderLabels(["Name", "Class", "Status"])
        table.verticalHeader().setVisible(False)
        table.resize(250, 50)
        table.resizeColumnToContents(0)
        return table


class InitializeDialog(QtGui.QDialog):

    def __init__(self, parent = None, auto_init=False, auto_close=False):
        super(InitializeDialog, self).__init__(parent)

        self.auto_init = auto_init
        self.auto_close = auto_close

        self.createGUI()

    def initialize(self):
        from ..connect import connect_initialize
        connect_initialize(self.widget, self.drivers,
                           dependencies=self.dependencies,
                           concurrent=self.concurrent)

    def exec_(self):
        self.show()

        if self.auto_init:
            self.initialize()

        return super().exec_()

    @staticmethod
    def from_flock(flock, auto_init=True, auto_close=True, concurrent=False, parent=None):
        dialog = InitializeDialog(parent, auto_init, auto_close)

        dialog.drivers = list(flock.values())
        dialog.dependencies = flock.dependencies
        dialog.concurrent = concurrent

        result = dialog.exec_()
        return result == QtGui.QDialog.Accepted

    def createGUI(self):

        self.widget = self._createGUITable()

        layout = QtGui.QVBoxLayout()

        if not self.auto_init:
            button = QtGui.QPushButton()
            button.setText('Initialize')
            button.setEnabled(True)
            button.clicked.connect(self.initialize)

            layout.addWidget(button)

        layout.addWidget(self.widget)

        # OK and Cancel buttons
        buttons = QtGui.QDialogButtonBox(
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel,
            1, self)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        self.setLayout(layout)

        self.resize(600, 300)
        self.setWindowTitle("Driver initialization")

    def _createGUITable(self):

        table = QtGui.QTableWidget(0, 3)
        table.setHorizontalHeaderLabels(["Name", "Class", "Status"])
        table.verticalHeader().setVisible(False)
        table.resize(250, 50)
        table.resizeColumnToContents(0)
        return table



# -*- coding: utf-8 -*-
"""
    lantz.qt.app
    ~~~~~~~~~~~~

    Implements base classes for graphical applications:
    - Backend
    - Frontend



    Backend
    -------

    Upon class definition
    - Declares multiple InstrumentSlots
    - Declares multiple BackendSlots

    Backends and instruments are instantiated by the user and connected between them by the user.


    Frontend
    --------

    Each frontend has a single backend.

    Upon class definition
    - Declares multiple Frontend / Frontends.using / Frontends.using_parent backend

    Frontends are instantiated and connected by Lantz


    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import os
import sys
import inspect
import collections
import threading

class HidingLock(object):
    def __init__(self, obj, lock=None):
        self.lock = lock or threading.RLock()
        self._obj = obj

    def __enter__(self):
        self.lock.acquire()
        return self._obj

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()



from pimpmyclass.mixins import LogMixin

from lantz.core.driver import Driver, Base
from lantz.core.flock import initialize_many, finalize_many
from lantz.core.helpers import MISSING

from .connect import connect_setup, connect_driver, connect_feat
from .widgets import DriverTestWidget, SetupTestWidget
from .objwrapper import QDriver
from .utils.qt import QtCore, QtGui, SuperQObject, MetaQObject
from .log import get_logger, LOGGER


ICON_FEDORA = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'assets', 'fedora.png')


class ThreadLogMixin:

    _PP_THREADS = HidingLock({})

    def log_current_thread(self, msg='Current thread is'):
        thread = self.thread()
        thid = int(thread.currentThreadId())
        thname = thread.objectName()

        with self._PP_THREADS as d:
            if thid in d:
                pp = d[thid]
            else:
                pp = d[thid] = len(d)

        self.log_debug(msg + ' %s (%d-%s)' % (thname, pp, thid))

    def moveToThread(self, thread):
        super().moveToThread(thread)
        self.log_current_thread('Moved to thread:')


class InstrumentSlot:

    def __str__(self):
        return '<InstrumentSlot>'

    def initialize(self):
        LOGGER.warning('The Instrument slot {} has not been assigned '
                       'to an actual instrument', self._slot_name )

    def finalize(self):
        pass


class FlockSlot:

    def __init__(self, *required_names):
        self.require_names = required_names

    def __str__(self):
        return '<FlockSlot>'

    def initialize(self):
        LOGGER.warning('The Flock slot {} has not been assigned '
                       'to an actual flock', self._slot_name )

    def finalize(self):
        pass


class BackendSlot:

    def __str__(self):
        return '<BackendSlot>'

    def initialize(self):
        LOGGER.warning('The Backend slot {} has not been assigned '
                       'to an actual instrument', self._slot_name )

    def finalize(self):
        pass


Front2Back = collections.namedtuple('Front2Back', 'frontend_class backend_name')


class Back2Back(collections.namedtuple('Back2Back', 'backend_class local_attribute foreign_attribute')):

    def link(self, local_name, foreign_name=None):
        return Back2Back(self.backend_class, local_name, foreign_name or local_name)


class _BackendType(MetaQObject):

    def create_instrument_property(cls, key):
        def getter(self):
            return self.instruments[key]
        def setter(self, instrument):
            if not isinstance(instrument, QtCore.QObject):
                raise ValueError("The object provided for InstrumentSlot '%s' is not subclass of QObject.\n"
                                 "Maybe you forgot to wrap your driver class?\n"
                                 "  from lantz.qt import wrap_driver_cls\n"
                                 "  QMyDriver = wrap_driver_cls(QMyDriver)\n" % key)
            instrument.name = key
            self.instruments[key] = instrument

        return property(getter, setter)

    def create_flock_property(cls, key):
        def getter(self):
            return self.flocks[key]
        def setter(self, flock):
            current = self.flocks[key]
            if set(current.required_names) <= set(flock.keys()):
                raise ValueError('The flock {} must have '
                                 'the following names (given {})'.format(key,
                                                                         set(current.required_names),
                                                                         set(flock.keys())))
            self.flocks[key] = flock

        return property(getter, setter)

    def create_backend_property(cls, key):
        def getter(self):
            return self.backends[key]
        def setter(self, backend):
            self.backends[key] = backend

        return property(getter, setter)

    def __init__(cls, classname, bases, class_dict):
        super().__init__(classname, bases, class_dict)

        cls.instruments = dict()
        cls.flocks = dict()
        cls.backends = dict()
        for key, value in class_dict.items():
            if isinstance(value, InstrumentSlot) or value is InstrumentSlot:
                if value is InstrumentSlot:
                    value = value()
                cls.instruments[key] = value
                cls.instruments[key]._slot_name = key
                setattr(cls, key, cls.create_instrument_property(key))
                LOGGER.debug('In {}, adding instrument named {} of type {}'.format(cls, key, value))

            elif isinstance(value, FlockSlot) or value is FlockSlot:
                if value is FlockSlot:
                    value = value()
                cls.flocks[key] = value
                cls.flocks[key]._slot_name = key
                setattr(cls, key, cls.create_flock_property(key))
                LOGGER.debug('In {}, adding flock named {} of type {}'.format(cls, key, value))

            elif value is BackendSlot:
                cls.backends[key] = value
                cls.backends[key]._slot_name = key
                setattr(cls, key, cls.create_backend_property(key))
                LOGGER.debug('In {}, adding backend named {} of type {}'.format(cls, key, value))

            else:
                pass
                #LOGGER.debug('In {}, unhandled attribute named {} = {}'.format(cls, key, value))

    def __str__(cls):
        return cls.__name__


class _FrontendType(MetaQObject):

    def create_frontend_property(cls, key):
        def getter(self):
            return self.frontends[key]
        def setter(self, frontend):
            self.frontends[key] = frontend

        return property(getter, setter)

    def __init__(cls, classname, bases, class_dict):
        super().__init__(classname, bases, class_dict)

        cls.frontends = dict()

        for key, value in class_dict.items():
            if isinstance(value, Front2Back) or hasattr(value, 'frontends'):
                cls.frontends[key] = value
                setattr(cls, key, cls.create_frontend_property(key))
                LOGGER.debug('{}, adding frontend named {} of type {}'.format(cls, key, value))

    def __str__(cls):
        return cls.__name__


class Frontend(LogMixin, ThreadLogMixin, QtGui.QMainWindow, metaclass=_FrontendType):

    logger_name = None
    _get_logger = get_logger

    frontends = {}

     # a declarative way to indicate the user interface file to use.
    gui = None

    # connect widgets to instruments using connect_setup
    auto_connect = False

    def __init__(self, parent=None, backend=None):
        super().__init__(parent)

        self._backend = None

        if self.logger_name is None:
            self.logger_name = 'lantz.qt.frontend.' + str(self)

        if self.gui:
            for cls in self.__class__.__mro__:
                if cls is object:
                    raise ValueError('{}: loading gui file {}, reached object parent'.format(self, self.gui))

                filename = os.path.dirname(inspect.getfile(cls))
                if isinstance(self.gui, tuple):
                    filename = os.path.join(filename, *self.gui)
                else:
                    filename = os.path.join(filename, self.gui)
                if os.path.exists(filename):
                    self.log_debug('loading gui file {}'.format(filename))
                    self.widget = QtGui.loadUi(filename)
                    if isinstance(self.widget, QtGui.QMainWindow):
                        QtGui.loadUi(filename, self)
                        break
                    self.setCentralWidget(self.widget)
                    break
            else:
                raise ValueError('{}: loading gui file {}'.format(self, self.gui))

        # Iterate over all frontend items in the current frontend
        # and instantiate each of them.
        # Note: a frontend declares which sub backend requires
        # but instantation is delegated to lantz

        for name, frontend in self.frontends.items():

            if isinstance(frontend, Front2Back):

                # The item is a frontend that requires a backend.

                cls = frontend.frontend_class
                if backend is None:
                    # If the current backend is None, then we cannot give
                    # anything to the sub frontend item
                    widget = cls()
                    self.log_debug('{} ({}) requires a backend but no backend defined'.format(name, cls))
                else:
                    # If the current backend exists then we get which of its attribute
                    # is the backend of the subfrontend
                    sub_backend_name = frontend.backend_name
                    if sub_backend_name is None:
                        # Is the same backend
                        widget = cls(backend=backend)
                        self.log_debug('{} ({}) connected to parent backend'.format(name, cls))
                    else:
                        # Is in an attribute
                        try:
                            widget = cls(backend=getattr(backend, sub_backend_name))
                        except AttributeError:
                            raise ValueError("{} ({}) requires a '{}' attribute which is not provided by {}".format(name, cls, sub_backend_name, backend))
                        self.log_debug('{} ({}) connected to backend.{}'.format(name, cls, sub_backend_name))
            else:
                # This backend does not declare a required backend
                self.log_debug('{} ({}) created'.format(name, frontend))
                widget = frontend()

            widget.setParent(self)
            setattr(self, name, widget)

        self.setupUi()
        self.backend = backend

        self.log_current_thread()

    def __str__(self):
        return self.__class__.__name__

    def setupUi(self):
        pass

    def connect_backend(self):
        pass

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, backend):
        if self._backend:
            self.log_debug('disconnecting backend: {}'.format(backend))
            self.disconnect(backend)

        self._backend = backend

        if backend:
            self.log_debug('connecting backend: {}'.format(backend))
            if self.auto_connect:
                connect_setup(self.widget, backend.instruments.values())

            self.connect_backend()

    @classmethod
    def using(cls, backend_name):
        return Front2Back(cls, backend_name)

    @classmethod
    def using_parent_backend(cls):
        return Front2Back(cls, backend_name=None)

    def connect_feat(self, widget, target, feat_name=None, feat_key=MISSING):
        return connect_feat(widget, target, feat_name, feat_key)


class Backend(Base, ThreadLogMixin, SuperQObject, metaclass=_BackendType):

    _observer_signal_init = lambda: QtCore.pyqtSignal(object, object)

    backends = None
    instruments = None
    flocks = None

    def __init__(self, parent=None, **instruments_and_backends):


        # First we check that the names provided do not collide with parent classes
        invalid = set(dir(Base)) | set(dir(SuperQObject)) | set(dir(_BackendType))
        for name in instruments_and_backends.keys():
            if name in invalid:
                raise ValueError('{} is an invalid instrument or backend name'
                                 ' as it collides with attribute of parent class'.format(name))

        if self.logger_name is None:
            self.logger_name = 'lantz.qt.backend.' + str(self)

        Base.__init__(self, self.logger_name)
        super().__init__(parent)

        inst_keys = set(self.instruments.keys())
        flo_keys = set(self.flocks.keys())
        be_keys = set(self.backends.keys())

        for key, value in instruments_and_backends.items():
            if key in inst_keys:
                inst_keys.remove(key)
                setattr(self, key, value)
            elif key in flo_keys:
                flo_keys.remove(key)
                self.flocks[key] = value
            elif key in be_keys:
                be_keys.remove(key)
                self.backends[key] = value
            else:
                raise ValueError('Cannot link argument. No InstrumentSlot, BackendSlot or FlockSlot named {}'.format(key))

        if inst_keys:
            raise ValueError('No value provided for InstrumentSlot {}'.format(inst_keys))

        if be_keys:
            raise ValueError('No value provided for BackendSlot {}'.format(be_keys))

        self.log_current_thread()

    def __str__(self):
        return self.__class__.__name__

    def initialize(self, register_finalizer=False):
        initialize_many(self.instruments.values(), register_finalizer=register_finalizer)

    def finalize(self):
        finalize_many(self.instruments.values())

    def __enter__(self):
        self.initialize(register_finalizer=False)
        return self

    def __exit__(self, *args):
        self.finalize()

    def invoke(self, func, *args, **kwargs):
        QtCore.QTimer.singleShot(0, lambda: func(*args, **kwargs))

    def moveToThread(self, thread):
        super().moveToThread(thread)
        for inst in self.instruments.values():
            inst.moveToThread(thread)
        for be in self.backends.values():
            be.moveToThread(thread)

def build_qapp(qapp_or_args=None, after_func=None):
    if isinstance(qapp_or_args, QtGui.QApplication):
        qapp = qapp_or_args
    else:
        qapp = QtGui.QApplication(qapp_or_args or [''])
        qapp.setWindowIcon(QtGui.QIcon(ICON_FEDORA))

    if after_func:
        qapp = after_func(qapp)

    return qapp


def start_gui_app(backend, frontend_class, qapp_or_args=None, after_qapp_creation=None):
    qapp = build_qapp(qapp_or_args, after_qapp_creation)

    QtCore.QThread.currentThread().setObjectName('main')

    background_thread = QtCore.QThread()
    background_thread.setObjectName('background')
    backend.moveToThread(background_thread)
    background_thread.start()

    frontend = frontend_class(backend=backend)
    frontend.show()

    qapp.aboutToQuit.connect(background_thread.quit)

    if sys.platform.startswith('darwin'):
        frontend.raise_()

    sys.exit(qapp.exec_())


def start_test_app(target, width=500, qapp_or_args=None, after_qapp_creation=None):
    """Start a single window test application with a form automatically
    generated for the driver.

    Parameters
    ----------
    target :
        a driver object or a collection of drivers.
    width :
        to be used as minimum width of the window. (Default value = 500)
    qapp_or_args :
        arguments to be passed to QApplication. (Default value = None)

    Returns
    -------

    """

    qapp = build_qapp(qapp_or_args, after_qapp_creation)

    if isinstance(target, (Driver, QDriver)):
        main = DriverTestWidget(None, target)
    else:
        main = SetupTestWidget(None, target)

    main.setMinimumWidth(width)
    main.setWindowTitle('Lantz Driver Test Panel')
    main.show()
    if sys.platform.startswith('darwin'):
        main.raise_()
    qapp.exec_()


def start_gui(ui_filename, drivers, qapp_or_args=None, after_qapp_creation=None):
    """Start a single window application with a form generated from
    a designer file.

    Parameters
    ----------
    ui_filename :
        the full path of a file generated with QtDesigner.
    drivers :
        a driver object or a collection of drivers.
    qapp_or_args :
        arguments to be passed to QApplication. (Default value = None)

    Returns
    -------

    """
    qapp = build_qapp(qapp_or_args, after_qapp_creation)

    main = QtGui.loadUi(ui_filename)

    if isinstance(drivers, Driver):
        connect_driver(main, drivers)
    else:
        connect_setup(main, drivers)

    main.show()
    if sys.platform.startswith('darwin'):
        main.raise_()
    qapp.exec_()


def start_frontend(frontend_class, qapp_or_args=None, after_qapp_creation=None):
    qapp = build_qapp(qapp_or_args, after_qapp_creation)

    frontend = frontend_class()
    frontend.show()

    if sys.platform.startswith('darwin'):
        frontend.raise_()

    sys.exit(qapp.exec_())




# -*- coding: utf-8 -*-
"""
    lantz.ui.connect
    ~~~~~~~~~~~~~~~~

    Implements UI widgets based on Qt widgets. To achieve functionality,
    instances of QtWidgets are patched.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import time

from lantz.core import initialize_many
from lantz.core import Driver
from lantz.core.driver import Base
from lantz.core.helpers import MISSING

from .log import LOGGER
from .utils.qt import QtCore, QtGui
from .widgets import WidgetMixin, ChildrenWidgets


def connect_feat(widget, target, feat_name=None, feat_key=MISSING):
    """Connect a feature from a given driver to a widget. Calling this
    function also patches the widget is necessary.
    
    If applied two times with the same widget, it will connect to the target
    provided in the second call. This behaviour can be useful to change the
    connection target without rebuilding the whole UI. Alternative, after
    connect has been called the first time, widget will have a property
    `lantz_target` that can be used to achieve the same thing.

    Parameters
    ----------
    widget : QtWidget
        widget instance.
    target : Driver
        driver instance.
    feat_name : str
        feature name. If None, connect using widget name. (Default value = None)
    feat_key : str
        For a DictFeat, this defines which key to show. (Default value = MISSING)

    """

    LOGGER.debug('Connecting {} to {}, {}, {}'.format(widget, target, feat_name, feat_key))

    if not isinstance(target, Base):
        raise TypeError('Connect target must be an instance of Driver or Backend, not {}'.format(target))

    if not feat_name:
        feat_name = widget.objectName()

    #: Reconnect
    if hasattr(widget, '_feat.name') and widget._feat.name == feat_name:
        widget.lantz_target = target
        return

    try:
        feat = target.feats[feat_name]
    except KeyError:
        try:
            feat = target.dictfeats[feat_name]
        except KeyError:
            raise KeyError("Cannot find feat/dictfeat '{}' in {}".format(feat_name, target))

    WidgetMixin.wrap(widget)
    widget.bind_feat(feat)
    widget.feat_key = feat_key

    widget.lantz_target = target


def connect_driver(parent, target, *, prefix='', sep='__'):
    """Connect all children widgets to their corresponding lantz feature
    matching by name. Non-matching names are ignored.

    Parameters
    ----------
    parent : QtWidget
        parent widget.
    target : driver
        the driver.
    prefix : str
        prefix to be prepended to the lantz feature (default = '')
    sep : str
        separator between prefix, name and suffix (Default value = '__')

    """

    LOGGER.debug('Connecting {} to {}, {}, {}'.format(parent, target, prefix, sep))

    ChildrenWidgets.patch(parent)

    if prefix:
        prefix += sep

    for name, _, wid in parent.widgets:
        if prefix and name.startswith(prefix):
            name = name[len(prefix):]
        if sep in name:
            name, _ = name.split(sep, 1)
        if name in target.feats:
            connect_feat(wid, target, name)


def connect_setup(parent, targets, *, prefix=None, sep='__'):
    """Connect all children widget to their corresponding

    Parameters
    ----------
    parent : QtWidget
        parent widget.
    targets : iterable of Driver
        iterable of drivers.
    prefix : str
        prefix to be prepended to the lantz feature name
        if None, the driver name will be used (default)
        if it is a dict, the driver name will be used to obtain
        he prefix.
    sep : str
         (Default value = '__')

    """

    LOGGER.debug('Connecting {} to {}, {}, {}'.format(parent, targets, prefix, sep))

    ChildrenWidgets.patch(parent)
    for target in targets:
        name = target.name
        if isinstance(prefix, dict):
            name = prefix[name]
        connect_driver(parent, target, prefix=name, sep=sep)


def connect_initialize_flock(widget, flock, register_finalizer=True,
                             initializing_msg='Initializing ...', initialized_msg='Initialized',
                             concurrent=True):
    """Initialize drivers while reporting the status in a QtWidget.

    Parameters
    ----------
    widget : QtWidget
        Qt Widget where the status information is going to be shown.
    flock : Flock object

    register_finalizer : bool
        register driver.finalize method to be called at python exit. (Default value = True)
    initializing_msg : str
        message to be displayed while initializing. (Default value = 'Initializing ...')
    initialized_msg : str
        message to be displayed after successful initialization. (Default value = 'Initialized')
    concurrent : bool
        indicates that drivers with satisfied dependencies
        should be initialized concurrently. (Default value = True)

    Returns
    -------
    type
        the QThread doing the initialization.

    """
    return connect_initialize(widget, flock.values(), register_finalizer,
                              initializing_msg, initialized_msg,
                              concurrent, flock.dependencies)


def connect_initialize(widget, drivers, register_finalizer=True,
                       initializing_msg='Initializing ...', initialized_msg='Initialized',
                       concurrent=True, dependencies=None):
    """Initialize drivers while reporting the status in a QtWidget.

    Parameters
    ----------
    widget : QtWidget
        Qt Widget where the status information is going to be shown.
    drivers : iterable of drivers
        iterable of drivers to initialize.
    register_finalizer : bool
        register driver.finalize method to be called at python exit. (Default value = True)
    initializing_msg : str
        message to be displayed while initializing. (Default value = 'Initializing ...')
    initialized_msg : str
        message to be displayed after successful initialization. (Default value = 'Initialized')
    concurrent : bool
        indicates that drivers with satisfied dependencies
        should be initialized concurrently. (Default value = True)
    dependencies : dict
        indicates which drivers depend on others to be initialized.
        each key is a driver name, and the corresponding
        value is an iterable with its dependencies. (Default value = None)

    Returns
    -------
    type
        the QThread doing the initialization.

    """
    timing = {}

    thread = QtCore.QThread()
    helper = InitializerHelper(drivers, register_finalizer, concurrent, dependencies)
    helper.moveToThread(thread)
    thread.helper = helper

    if isinstance(widget, QtGui.QTableWidget):
        def _initializing(driver):
            timing[driver] = time.perf_counter()
            row = drivers.index(driver)
            widget.setItem(row, 2, QtGui.QTableWidgetItem(initializing_msg))

        def _initialized(driver):
            delta = time.perf_counter() - timing[driver]
            row = drivers.index(driver)
            widget.setItem(row, 2, QtGui.QTableWidgetItem(initialized_msg + ' ({:.1f} sec)'.format(delta)))

        def _exception(driver, e):
            delta = time.perf_counter() - timing[driver]
            row = drivers.index(driver)
            widget.setItem(row, 2, QtGui.QTableWidgetItem('{} ({:.1f} sec)'.format(e, delta)))

        def _done(duration):
            widget.setItem(len(drivers), 2, QtGui.QTableWidgetItem('{:.1f} sec'.format(duration)))
            thread.quit()

        widget.clearContents()
        widget.setRowCount(len(drivers) + 1)
        for ndx, onedriver in enumerate(drivers):
            widget.setItem(ndx, 0, QtGui.QTableWidgetItem(onedriver.name))
            widget.setItem(ndx, 1, QtGui.QTableWidgetItem(onedriver.__class__.__name__))
            widget.setItem(ndx, 2, QtGui.QTableWidgetItem(''))

        widget.resizeColumnToContents(0)
        widget.horizontalHeader().setStretchLastSection(True)

    elif isinstance(widget, QtGui.QLineEdit):
        def _initializing(driver):
            timing[driver] = time.perf_counter()
            widget.setText('{} ({}) > {}'.format(driver.name, driver.__class__.__name__,
                                                 initializing_msg))

        def _initialized(driver):
            delta = time.perf_counter() - timing[driver]
            widget.setText('{} ({}) > {} ({:.1f} sec)'.format(driver.name, driver.__class__.__name__,
                                                              initialized_msg, delta))

        def _exception(driver, e):
            delta = time.perf_counter() - timing[driver]
            widget.setText('{} ({}) > {} ({:.1f} sec)'.format(driver.name, driver.__class__.__name__,
                                                              e, delta))

        def _done(duration):
            widget.setText('Initialized in {:.1f} sec'.format(duration))
            thread.quit()

        widget.setReadOnly(True)

    elif isinstance(widget, QtGui.QTextEdit):

        def _initializing(driver):
            timing[driver] = time.perf_counter()
            widget.append('{} ({}) > {}'.format(driver.name, driver.__class__.__name__,
                                                initializing_msg))

        def _initialized(driver):
            delta = time.perf_counter() - timing[driver]
            widget.append('{} ({}) > {} ({:.1f} sec)'.format(driver.name, driver.__class__.__name__,
                                                             initialized_msg, delta))

        def _exception(driver, e):
            delta = time.perf_counter() - timing[driver]
            widget.append('{} ({}) > {} ({:.1f} sec)'.format(driver.name, driver.__class__.__name__,
                                                             e, delta))

        def _done(duration):
            widget.append('Initialized in {:.1f} sec'.format(duration))
            thread.quit()

        widget.setReadOnly(True)

    else:
        raise TypeError('Unknown widget type {}.'.format(type(widget)))

    thread.started.connect(helper.process)
    helper.initializing.connect(_initializing)
    helper.initialized.connect(_initialized)
    helper.exception.connect(_exception)
    helper.finished.connect(_done)

    thread.start()
    return thread


class InitializerHelper(QtCore.QObject):

    initializing = QtCore.Signal(object)
    initialized = QtCore.Signal(object)
    exception = QtCore.Signal(object, object)
    finished = QtCore.Signal(float)

    def __init__(self, drivers, register_finalizer, parallel, dependencies):
        super().__init__()
        self.drivers = drivers
        self.register_finalizer = register_finalizer
        self.parallel = parallel
        self.dependencies = dependencies

    def process(self):
        start = time.perf_counter()
        initialize_many(drivers=self.drivers, register_finalizer=self.register_finalizer,
                        on_initializing=self.on_initializing,
                        on_initialized=self.on_initialized,
                        on_exception=self.on_exception,
                        concurrent=self.parallel,
                        dependencies=self.dependencies)
        self.finished.emit(time.perf_counter() - start)

    def on_initializing(self, driver):
        self.initializing.emit(driver)

    def on_initialized(self, driver):
        self.initialized.emit(driver)

    def on_exception(self, driver, ex):
        self.exception.emit(driver, ex)



# -*- coding: utf-8 -*-
"""
    lantz.qt.log
    ~~~~~~~~~~~~

    Implements base class for graphical applications.

    :copyright: 2018 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from lantz.core.log import get_logger

LOGGER = get_logger('lantz.qt')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

import codecs

def read(filename):
    return codecs.open(filename, encoding='utf-8').read()


long_description = '\n\n'.join([read('README'),
                                read('AUTHORS'),
                                read('CHANGES')])

__doc__ = long_description

setup(name='lantz-qt',
      version='0.5',
      license='BSD',
      description='Instrumentation framework',
      long_description=long_description,
      keywords='measurement control instrumentation science',
      author='Hernan E. Grecco',
      author_email='hernan.grecco@gmail.com',
      url='https://github.com/lantzproject',
      install_requires=['lantz-core>=0.5.dev1',
                        ],
      include_package_data=True,
      packages=['lantz.qt',
                'lantz.qt.blocks',
                'lantz.qt.utils',
                'lantz.qt.widgets'],
      zip_safe=False,
      platforms='any',
      entry_points={
          'console_scripts': [
              'lantz-qtdemo = lantz.qt.__main__:main'
          ],
          'lantz_subcommands': [
              'qtdemo = lantz.qt.__main__:main'
          ]
      },
      classifiers=[
           'Development Status :: 4 - Beta',
           'Intended Audience :: Developers',
           'Intended Audience :: Science/Research',
           'License :: OSI Approved :: BSD License',
           'Operating System :: MacOS :: MacOS X',
           'Operating System :: Microsoft :: Windows',
           'Operating System :: POSIX',
           'Programming Language :: Python',
           'Programming Language :: Python :: 3.6',
           'Topic :: Scientific/Engineering',
           'Topic :: Software Development :: Libraries'
      ],
)



# -*- coding: utf-8 -*-
"""
    lantz.ui.loop
    ~~~~~~~~~~~~~

    A Loop backend and frontend.

    :copyright: 2015 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import time
import math
from enum import IntEnum

from ..utils.qt import QtCore, QtGui
from ..app import Frontend, Backend, start_gui_app


class StopMode(IntEnum):
    Continuous = 0
    Duration = 1
    Iterations = 2
    IterationsTimeOut = 3


class Loop(Backend):
    """The Loop backend allows you to execute task periodically.
    
    Usage:
    
        from lantz.ui.blocks import Loop, LoopUi
    
        def measure(counter, iterations, overrun):
            print(counter, iterations, overrun)
            data = osci.measure()
            print(data)
    
        app = Loop()
    
        app.body = measure
    
        start_gui_app(app, LoopUi)

    Parameters
    ----------

    Returns
    -------

    """

    #: Signal emitted before starting a new iteration
    #: Parameters: loop counter, iterations, overrun
    iteration = QtCore.Signal(int, int, bool)

    #: Signal emitted when the loop finished.
    #: The parameter is used to inform if the loop was canceled.
    loop_done = QtCore.Signal(bool)

    #: The function to be called. It requires three parameters.
    #:   counter - the iteration number
    #:   iterations - total number of iterations
    #:   overrun - a boolean indicating if the time required for the operation
    #:            is longer than the interval.
    #: :type: (int, int, bool) -> None
    body = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._active = False
        self._internal_func = None

    def stop(self):
        """Request the scanning to be stop.
        Will stop when the current iteration is finished.

        Parameters
        ----------

        Returns
        -------

        """
        self._active = False

    def start(self, body, interval=0, iterations=0, timeout=0):
        """Request the scanning to be started.

        Parameters
        ----------
        body :
            function to be called at each iteration.
            If None, the class body will be used.
        interval :
            interval between starts of the iteration.
            If the body takes too long, the iteration will
            be as fast as possible and the overrun flag will be True (Default value = 0)
        iterations :
            number of iterations (Default value = 0)
        timeout :
            total time in seconds that the scanning will take.
            If overdue, the scanning will be stopped.
            If 0, there is no timeout. (Default value = 0)

        Returns
        -------

        """
        self._active = True
        body = body or self.body

        def internal(counter, overrun=False, schedule=QtCore.QTimer.singleShot):
            if not self._active:
                self.loop_done.emit(True)
                return

            st = time.time()
            self.iteration.emit(counter, iterations, overrun)
            body(counter, iterations, overrun)

            if iterations and counter + 1 == iterations:
                self._active = False
                self.loop_done.emit(False)
                return
            elif not self._active:
                self.loop_done.emit(True)
                return

            sleep = interval - (time.time() - st)
            schedule(sleep * 1000 if sleep > 0 else 0,
                     lambda: self._internal_func(counter + 1, sleep < 0))

        self._internal_func = internal
        if timeout:
            QtCore.QTimer.singleShot(timeout * 1000, self.stop)
        QtCore.QTimer.singleShot(0, lambda: self._internal_func(0))


class LoopUi(Frontend):
    """The Loop frontend provides a GUI for the Rich Backend"""

    gui = 'loop.ui'

    auto_connect = False

    #: Signal emitted when a start is requested.
    #: The parameters are None, interval, iterations, duration
    request_start = QtCore.Signal(object, object, object, object)

    #: Signal emitted when a stop is requested.
    request_stop = QtCore.Signal()

    def connect_backend(self):
        super().connect_backend()

        self.widget.start_stop.clicked.connect(self.on_start_stop_clicked)
        self.widget.mode.currentIndexChanged.connect(self.on_mode_changed)
        self.widget.iterations.valueChanged.connect(self.recalculate)
        self.widget.duration.valueChanged.connect(self.recalculate)
        self.widget.interval.valueChanged.connect(self.recalculate)
        self.widget.progress_bar.setValue(0)

        self._ok_palette = QtGui.QPalette(self.widget.progress_bar.palette())
        self._overrun_palette = QtGui.QPalette(self.widget.progress_bar.palette())
        self._overrun_palette.setColor(QtGui.QPalette.Highlight,
                                       QtGui.QColor(QtCore.Qt.red))

        self.backend.iteration.connect(self.on_iteration)
        self.backend.loop_done.connect(self.on_loop_done)

        self.request_start.connect(self.backend.start)
        self.request_stop.connect(self.backend.stop)

    def on_start_stop_clicked(self, value=None):
        if self.backend._active:
            self.widget.start_stop.setText('...')
            self.widget.start_stop.setEnabled(False)
            self.request_stop.emit()
            return

        self.widget.start_stop.setText('Stop')
        self.widget.start_stop.setChecked(True)

        mode = self.widget.mode.currentIndex()
        interval, iterations, duration = [getattr(self.widget, name).value()
                                          for name in 'interval iterations duration'.split()]
        if mode == StopMode.Continuous:
            self.request_start.emit(None, interval, 0, 0)
        elif mode == StopMode.Iterations:
            self.request_start.emit(None, interval, iterations, 0)
        elif mode == StopMode.Duration:
            self.request_start.emit(None, interval, 0, duration)
        elif mode == StopMode.IterationsTimeOut:
            self.request_start.emit(None, interval, iterations, duration)

    def recalculate(self, *args):
        mode = self.widget.mode.currentIndex()
        if mode == StopMode.Duration:
            iterations = self.widget.duration.value() / self.widget.interval.value()
            self.widget.iterations.setValue(math.ceil(iterations))
        elif mode == StopMode.Iterations:
            self.widget.duration.setValue(self.widget.iterations.value() * self.widget.interval.value())

    def on_iteration(self, counter, iterations, overrun):
        pbar = self.widget.progress_bar

        if not counter:
            if iterations:
                pbar.setMaximum(iterations + 1)
            else:
                pbar.setMaximum(0)

        if iterations:
            pbar.setValue(counter + 1)

        if overrun:
            pbar.setPalette(self._overrun_palette)
        else:
            pbar.setPalette(self._ok_palette)

    def on_mode_changed(self, new_index):
        if new_index == StopMode.Continuous:
            self.widget.duration.setEnabled(False)
            self.widget.iterations.setEnabled(False)
        elif new_index == StopMode.Duration:
            self.widget.duration.setEnabled(True)
            self.widget.iterations.setEnabled(True)
            self.widget.duration.setReadOnly(False)
            self.widget.iterations.setReadOnly(True)
        elif new_index == StopMode.Iterations:
            self.widget.duration.setEnabled(True)
            self.widget.iterations.setEnabled(True)
            self.widget.duration.setReadOnly(True)
            self.widget.iterations.setReadOnly(False)
        elif new_index == StopMode.IterationsTimeOut:
            self.widget.duration.setEnabled(True)
            self.widget.iterations.setEnabled(True)
            self.widget.duration.setReadOnly(False)
            self.widget.iterations.setReadOnly(False)
        self.recalculate()

    def on_loop_done(self, cancelled):
        self.widget.start_stop.setText('Start')
        self.widget.start_stop.setEnabled(True)
        self.widget.start_stop.setChecked(False)
        if self.widget.progress_bar.maximum():
            self.widget.progress_bar.setValue(self.widget.progress_bar.maximum())
        else:
            self.widget.progress_bar.setMaximum(1)


if __name__ == '__main__':
    def func(current, total, overrun):
        print('func', current, total, overrun)
    app = Loop()
    app.body = func
    start_gui_app(app, LoopUi)
